## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . xmlns import xmlns
## END_IMPORT

# ====================================================================
class Element:
    BUILD  = 'VASSAL.build.'
    MODULE = BUILD  + 'module.'
    WIDGET = BUILD  + 'widget.'
    FOLDER = MODULE + 'folder.'
    MAP    = MODULE + 'map.'    
    PICKER = MAP    + 'boardPicker.'
    BOARD  = PICKER + 'board.'
    known_tags   = {}
    unique_attrs = {}
    
    def __init__(self,parent,tag,node=None,**kwargs):
        '''Create a new element

        Parameters
        ----------
        parent : Element
            Parent element to add this element to
        tag : str
            Element tag
        node : xml.dom.Node
            If not None, then read attributes from that. Otherwise
            set elements according to kwargs
        kwargs : dict
            Attribute keys and values.  Only used if node is None
        '''
        #from xml.dom.minidom import Document
        
        if parent is not None:
            self._tag  = tag                
            self._root = (parent if isinstance(parent,xmlns.Document) else
                          parent._root)
            self._node = (node if node is not None else
                          parent.addNode(tag,**kwargs))
        else:
            self._root = None
            self._node = None
            self._tag  = None

    # ----------------------------------------------------------------
    @classmethod
    def _make_unique(cls,tag,*values):
        return tag + ('_'+'_'.join(values) if len(values)>0 else '')
    
    def _unique(self):
        uattr = Element.unique_attrs.get(self._tag,[])
        return self._make_unique(self._tag,
                                 *[self.getAttribute(a) for a in uattr])
    
    def __hash__(self):
        return hash(self._unique())
    
    def __eq__(self,other):
        '''Equality comparison - check if to elements are the same.

        This is based on the tag of the elements first, then on the
        attributes.  However, not all attributes should be compared
        for equality - only thise that are meant to be unique.

        '''
        #print(f'Compare {self} to {other} ({self._tag},{other._tag})')
        if not isinstance(other,Element):
            return False
        return self.__hash__() == other.__hash__()
        #if self._tag != other._tag:
        #    return False

        # to be done
        #
        # uattr = Element.unique_attrs[self._tag]
        # attr  = self .getAttributes()
        # oattr = other.getAttributes()
        
        return True

        
    # ----------------------------------------------------------------
    # Attributes
    def __contains__(self,key):
        '''Check if element has attribute key'''        
        return self.hasAttribute(key)
    
    def __getitem__(self,key):
        '''Get attribute key value'''
        return self.getAttribute(key)

    def __setitem__(self,key,value):
        '''Set attribute key value'''
        self.setAttribute(key,value)

    def hasAttribute(self,k):
        '''Check if element has attribute '''
        return self._node.hasAttribute(k)

    def getAttribute(self,k):
        '''Get attribute key value'''
        return self._node.getAttribute(k)
        
    def setAttribute(self,k,v):
        '''Set attribute key value'''
        self._node.setAttribute(k,str(v).lower()
                                if isinstance(v,bool) else str(v))
        
    def setAttributes(self,**kwargs):
        '''Set attributes to dictionary key and value'''
        for k,v in kwargs.items():
            self.setAttribute(k,v)

    def getAttributes(self):
        '''Get attributes as dict'''
        return self._node.attributes

    # ----------------------------------------------------------------
    # Plain nodes
    def getChildren(self):
        '''Get child nodes (xml.dom.Node)'''
        return self._node.childNodes

    # ----------------------------------------------------------------
    # Getters
    #
    # First generics 
    def getAsDict(self,tag='',key=None,enable=True):
        '''Get elements with a specific tag as a dictionary
        where the key is given by attribute key'''
        cont = self._node.getElementsByTagName(tag)
        if not enable or key is None:
            return cont

        return {e.getAttribute(key): e for e in cont}

    def getAsOne(self,tag='',single=True):
        '''Get elements with a specific tag, as a list.
        If single is true, then assume we only have one such
        child element, or fail.'''
        cont = self._node.getElementsByTagName(tag)
        if single and len(cont) != 1:
            return None
        return cont
    
    def getElementsByKey(self,cls,key='',asdict=True):
        '''Get elments of a specific class as a dictionary,
        where the key is set by the key attribute.'''
        cont = self.getAsDict(cls.TAG,key,asdict)
        if cont is None: return None
        
        if not asdict: return [cls(self,node=n) for n in cont]

        return {k : cls(self,node=n) for k, n in cont.items()}

    def getAllElements(self,cls,single=True):
        '''Get elements with a specific tag, as a list.  If single is
        true, then assume we only have one such child element, or
        fail.

        If `cls` is None, then return _all_ child elements. 

        '''
        #from xml.dom.minidom import Text, Element as XMLElement
        if cls is None:
            ret = []
            for node in self.getChildren():
                if isinstance(node,xmlns.Text):
                    continue
                
                if not hasattr(node,'tagName'):
                    print(f'Do not know how to deal with {type(node)}')
                    continue

                tag = node.tagName
                cls = Element.getTagClass(tag)
                if cls is None:
                    raise RuntimeError(f'No class reflection of tag {tag}')
                
                ret.append(cls(self,node=node))

            return ret
                           
        cont = self.getAsOne(cls.TAG,single=single)
        if cont is None: return None
        return [cls(self,node=n) for n in cont]

    def getSpecificElements(self,cls,key,*names,asdict=True):
        '''Get all elements of specific class and that has the
        attribute key, and the attribute value is in names

        '''
        cont = self.getAsOne(cls.TAG,single=False)
        cand = [cls(self,node=n) for n in cont
                if n.getAttribute(key) in names]
        if asdict:
            return {c[key] : c for c in cand}
        return cand
    
    def getParent(self,cls=None,checkTag=True):
        if self._node.parentNode is None:
            return None
        if cls is None:
            cls = self.getTagClass(self._node.parentNode.tagName)
            checkTag = False
        if cls is None:
            return None
        if checkTag and self._node.parentNode.tagName != cls.TAG:
            return None
        return cls(self,node=self._node.parentNode)

    def getParentOfClass(self,cls):
        '''Searches back until we find the parent with the right
        class, or none
        '''
        try:
            iter(cls)
        except:
            cls = [cls]
        t = {c.TAG: c for c in cls}
        p = self._node.parentNode
        while p is not None:
            c = t.get(p.tagName,None)
            if c is not None: return c(self,node=p)
            p = p.parentNode
        return None

    @classmethod
    def getTagClass(cls,tag):
        '''Get class corresponding to the tag'''
        # if tag not in cls.known_tags: return None;
        # Older VASSAL may have funny tag-names
        return cls.known_tags.get(tag,None)
        
    # ----------------------------------------------------------------
    # Adders
    def addNode(self,tag,**attr):
        '''Add a note to this element

        Parameters
        ----------
        tag : str
            Node tag name
        attr : dict
            Attributes to set
        '''
        e = self._root.createElement(tag)
        if self._node: self._node.appendChild(e)

        for k, v in attr.items():
            e.setAttribute(k,str(v).lower() if isinstance(v,bool) else str(v))

        return e

    def addText(self,text):
        '''Add a text child node to an element'''
        t = self._root.createTextNode(text)
        self._node.appendChild(t)
        return t

    def hasText(self):
        return self._node.firstChild is not None and \
            self._node.firstChild.nodeType == self._node.firstChild.TEXT_NODE
        
    def getText(self):
        '''Get contained text node content'''
        if self._node.firstChild is None or \
           self._node.firstChild.nodeType != self._node.firstChild.TEXT_NODE:
            return ''
        return self._node.firstChild.nodeValue

    def setText(self,txt):
        '''Set contained text node content'''
        if self._node.firstChild is None or \
           self._node.firstChild.nodeType != self._node.firstChild.TEXT_NODE:
            return 
        self._node.firstChild.nodeValue = txt
    

    def add(self,cls,**kwargs):
        '''Add an element and return wrapped in cls object'''
        return cls(self,node=None,**kwargs)

    def append(self,elem):
        '''Append and element'''
        if self._node.appendChild(elem._node):
            return elem
        return False

    # ----------------------------------------------------------------
    def remove(self,elem):
        '''Remove an element'''
        try:
            self._node.removeChild(elem._node)
        except:
            return None
        return elem
    # ----------------------------------------------------------------
    def insertBefore(self,toadd,ref):
        '''Insert an element before another element'''
        try:
            self._node.insertBefore(toadd._node,ref._node)
        except:
            return None
        return toadd

    # ----------------------------------------------------------------
    def print(self,file=None,recursive=False,indent=''):
        '''Print this element, and possibly its child elements.

        If `file` is None, then print to stdout.  If `recursive` is
        `True`, then also print child elements.  If `recursive` is an
        integer, then print this many deep levels of child elements.

        '''
        if file is None:
            from sys import stdout
            file = stdout

        from io import StringIO
        from textwrap import indent as i

        stream = StringIO()
        
        print(f'Element TAG={self.TAG} CLS={self.__class__.__name__}',
              file=stream)
        attrs = self.getAttributes()
        #print(type(attrs))
        ln    = max([len(n) for n in attrs.keys()]+[0])
        for name,value in attrs.items():
            print(f' {name:{ln}s}: {value}',file=stream)

        if isinstance(recursive,bool):
            recursive = 1024 if recursive else 0# Large number
            
        if recursive > 1:
            for child in self.getAllElements(cls=None):
                child.print(file=stream,
                            recursive=recursive-1,
                            indent='  ')
        else:
            n = len(self.getChildren())
            if n > 0:
                print(f'  {n} child elements',file=stream)

        print(i(stream.getvalue(),indent).rstrip(),file=file)
            
            
        
# --------------------------------------------------------------------
class DummyElement(Element):
    def __init__(self,parent,node=None,**kwargs):
        '''A dummy element we can use to select elements of different
        classes

        '''  
        super(DummyElement,self).__init__(parent,'Dummy',node=node)

# --------------------------------------------------------------------
class ToolbarElement(Element):
    def __init__(self,
                 parent,
                 tag,
                 node         = None,
                 name         = '', # Toolbar element name
                 tooltip      = '', # Tool tip
                 text         = '', # Button text
                 icon         = '', # Button icon,
                 hotkey       = '', # Named key or key stroke
                 canDisable   = False,
                 propertyGate = '',
                 disabledIcon = '',
                 **kwargs):
        '''Base class for toolbar elements.

        Parameters
        ----------
        parent : Element
            Parent element if any
        tag : str
            Element tag
        node : XMLNode
            Possible node - when reading back
        name : str
            Name of element (user reminder).  If not set, and tooltip is set,
            set to tooltip
        toolttip : str        
            Tool tip when hovering. If not set, and name is set, then
            use name as tooltip.
        text : str
            Text of button
        icon : str
            Image path for button image
        hotkey : str
            Named key or key-sequence
        canDisable : bool
            If true, then the element can be disabled
        propertyGate : str        
            Name of a global property.  When this property is `true`,
            then this element is _disabled_.  Note that this _must_ be
            the name of a property - it cannot be a BeanShell
            expression.
        disabledIcon : str
            Path to image to use when the element is disabled.
        kwargs : dict
            Other attributes to set on the element
        '''
        if name == '' and tooltip != '': name    = tooltip
        if name != '' and tooltip == '': tooltip = name

        # Build arguments for super class 
        args = {
            'node':         node,
            'name':         name,
            'icon':         icon,
            'tooltip':      tooltip,
            'hotkey':       hotkey,
            'canDisable':   canDisable,
            'propertyGate': propertyGate,
            'disabledIcon': disabledIcon }
        bt = kwargs.pop('buttonText',None)
        # If the element expects buttonText attribute, then do not set
        # the text attribute - some elements interpret that as a
        # legacy name attribute,
        if bt is not None:
            args['buttonText'] = bt
        else:
            args['text']       = text
        args.update(kwargs)

        super(ToolbarElement,self).__init__(parent,
                                            tag,
                                            **args)
        # print('Attributes\n','\n'.join([f'- {k}="{v}"' for k,v in self._node.attributes.items()]))
        
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
# 
# 
class ChangePropertyTrait(Trait):
    DIRECT    = 'P'
    INCREMENT = 'I'
    PROMPT    = 'R'
    LIST      = 'E'
    def __init__(self,
                 *commands,
                 numeric     = False,
                 min         = 0,
                 max         = 100,
                 wrap        = False):
        '''Base class for property (piece or global) change traits.
        
           Encodes constraints and commands.
        '''
        # assert name is not None and len(name) > 0, \
        #     'No name specified for ChangePropertyTriat'
        super(ChangePropertyTrait,self).__init__()
        self._constraints = self.encodeConstraints(numeric,wrap,min,max)
        self._commands    = self.encodeCommands(commands)

    def encodeConstraints(self,numeric,wrap,min,max):
        isnum             = f'{numeric}'.lower()
        iswrap            = f'{wrap}'.lower()
        return f'{isnum},{min},{max},{iswrap}'

    def decodeConstraints(self,constraints):
        f = Trait.decodeKeys(constraints)
        return f[0]=='true',f[3]=='true',int(f[1]),int(f[2])
    
    def encodeCommands(self,commands):
        cmds              = []
        for cmd in commands:
            # print(cmd)
            com = cmd[0] + ':' + cmd[1].replace(',',r'\,') + ':' + cmd[2]
            if cmd[2] == self.DIRECT:
                com += r'\,'+cmd[3].replace(',',r'\\,').replace(':',r'\:')
            elif cmd[2] == self.INCREMENT:
                com += r'\,'+cmd[3].replace(',',r'\\,').replace(':',r'\:')
            cmds.append(com)
        # print(cmds)
        return ','.join(cmds)

    def decodeCommands(self,commands):
        cmds = Trait.decodeKeys(commands)
        ret  = []
        for cmd in cmds:
            parts = Trait.decodeKeys(cmd,':')
            # print('parts',parts)
            if parts[-1][0] == self.DIRECT:
                parts = parts[:-1]+Trait.decodeKeys(parts[-1],',')
            if parts[-1][0] == self.INCREMENT:
                parts = parts[:-1]+Trait.decodeKeys(parts[-1],',')
            ret.append(parts)
        # print(commands,parts)
        return ret
    
    def getCommands(self):
        return self.decodeCommands(self['commands'])

    def setCommands(self,commands):
        self['commands'] = self.encodeCommands(commands)
        
    def check(self):
        assert len(self['name']) > 0,\
            f'No name given for ChangePropertyTrait'
        
        
# --------------------------------------------------------------------
class DynamicPropertyTrait(ChangePropertyTrait):
    ID = 'PROP'
    def __init__(self,
                 *commands,
                 name        = '',
                 value       = 0,
                 numeric     = False,
                 min         = 0,
                 max         = 100,
                 wrap        = False,
                 description = ''):
        '''Commands are

            - menu
            - key
            - Type (only 'P' for now)
            - Expression
        '''
        super(DynamicPropertyTrait,self).__init__(*commands,
                                                  numeric = numeric,
                                                  min     = min,
                                                  max     = max,
                                                  wrap    = wrap)
        # print(commands,'Name',name)
        self.setType(name        = name,
                     constraints = self._constraints,
                     commands    = self._commands,
                     description = description)
        self.setState(value=value)

    
Trait.known_traits.append(DynamicPropertyTrait)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . xmlns import xmlns
## END_IMPORT

# --------------------------------------------------------------------
class Data(Element):
    TAG = 'data'
    def __init__(self,doc,node=None,version='1'):
        super(Data,self).__init__(doc,self.TAG,node=node,version=version)
        
    def addVersion(self,**kwargs):
        '''Add a `Version` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Version
            The added element
        '''
        return self.add(Version,**kwargs)
    def addVASSALVersion(self,**kwargs):
        '''Add a `VASSALVersion` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : VASSALVersion
            The added element
        '''
        return self.add(VASSALVersion,**kwargs)
    def addName(self,**kwargs):
        '''Add a `Name` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Name
            The added element
        '''
        return self.add(Name,**kwargs)
    def addDescription(self,**kwargs):
        '''Add a `Description` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Description
            The added element
        '''
        return self.add(Description,**kwargs)
    def addDateSaved(self,**kwargs):
        '''Add a `DateSaved` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : DateSaved
            The added element
        '''
        return self.add(DateSaved,**kwargs)
    def getVersion(self,single=True):
        '''Get all or a sole `Version` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Version` child, otherwise fail.
            If `False` return all `Version` children in this element
        
        Returns
        -------
        children : list
            List of `Version` children (even if `single=True`)
        '''
        return self.getAllElements(Version,single=single)
    def getVASSALVersion(self,single=True):
        '''Get all or a sole `VASSALVersion` element(s) from this

        Parameters
        ----------
        single : bool        
            If `True`, there can be only one `VASSALVersion` child,
            otherwise fail.  If `False` return all `VASSALVersion`
            children in this element
        
        Returns
        -------
        children : list
            List of `VASSALVersion` children (even if `single=True`)

        '''
        return self.getAllElements(VASSALVersion,single=single)
    def getName(self,single=True):
        return self.getAllElements(Name,single=single)
    def getDescription(self,single=True):
        '''Get all or a sole `Description` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Description` child,
            otherwise fail.  If `False` return all `De` children in
            this element
        
        Returns
        -------
        children : list
            List of `De` children (even if `single=True`)

        '''
        return self.getAllElements(Description,single=single)
    def getDateSaved(self,single=True):
        '''Get all or a sole `DateSaved` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `DateSaved` child, otherwise fail.
            If `False` return all `DateSaved` children in this element
        
        Returns
        -------
        children : list
            List of `DateSaved` children (even if `single=True`)
        '''
        return self.getAllElements(DateSaved,single=single)
    
registerElement(Data)

# --------------------------------------------------------------------
class DataElement(Element):
    def __init__(self,data,tag,node=None,**kwargs):
        super(DataElement,self).__init__(data,tag,node=node,**kwargs)

    def getData(self):
        return self.getParent(Data)

# --------------------------------------------------------------------
class Version(DataElement):
    TAG = 'version'
    def __init__(self,data,node=None,version=''):
        super(Version,self).__init__(data,self.TAG,node=node)
        if node is None:
            self.addText(version)

registerElement(Version)

# --------------------------------------------------------------------
class Extra1(DataElement):
    TAG = 'extra1'
    def __init__(self,data,node=None,extra=''):
        super(Extra1,self).__init__(data,self.TAG,node=node)
        if node is None:
            self.addText(extra)

registerElement(Extra1)

# --------------------------------------------------------------------
class Extra2(DataElement):
    TAG = 'extra2'
    def __init__(self,data,node=None,extra=''):
        super(Extra2,self).__init__(data,self.TAG,node=node)
        if node is None:
            self.addText(extra)

registerElement(Extra2)

# --------------------------------------------------------------------
class VASSALVersion(DataElement):
    TAG = 'VassalVersion'
    def __init__(self,data,node=None,version='3.6.7'):
        super(VASSALVersion,self).__init__(data,self.TAG,node=node)
        if node is None:
            self.addText(version)

registerElement(VASSALVersion)

# --------------------------------------------------------------------
class Name(DataElement):
    TAG = 'name'
    def __init__(self,data,node=None,name=''):
        super(Name,self).__init__(data,self.TAG,node=node)
        if node is None:
            self.addText(name)
            
registerElement(Name)

# --------------------------------------------------------------------
class Description(DataElement):
    TAG = 'description'
    def __init__(self,data,node=None,description=''):
        super(Description,self).__init__(data,self.TAG,node=node)
        if node is None:
            self.addText(description)

registerElement(Description)

# --------------------------------------------------------------------
class DateSaved(DataElement):
    TAG = 'dateSaved'
    def __init__(self,data,node=None,milisecondsSinceEpoch=-1):
        super(DateSaved,self).__init__(data,self.TAG,node=node)
        if node is None:
            from time import time
            s = f'{int(time()*1000)}' if milisecondsSinceEpoch < 0 else \
                str(milisecondsSinceEpoch)
            self.addText(s)
            
registerElement(DateSaved)

# --------------------------------------------------------------------
class ModuleData(Element):

    def __init__(self,root=None):
        '''Construct from a DOM object, if given, otherwise make new'''
        #from xml.dom.minidom import Document
        super(ModuleData,self).__init__(None,'',None)
        
        self._root = root
        self._tag  = 'moduledata'
        if self._root is None:
            self._root = xmlns.Document()

        self._node = self._root

    def addData(self,**kwargs):
        '''Add a `Data` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Data
            The added element
        '''
        return Data(self,**kwargs)

    def getData(self):
        return Data(self,
                    node=self._root.getElementsByTagName(Data.TAG)[0])

    def encode(self):
        return self._root.toprettyxml(indent=' ',
                                      encoding="UTF-8",
                                      standalone=False)


#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . gameelements import GameElement
## END_IMPORT

# --------------------------------------------------------------------
class PlayerRoster(GameElement):
    TAG = Element.MODULE+'PlayerRoster'
    def __init__(self,doc,node=None,buttonKeyStroke='',
               buttonText='Retire',
               buttonToolTip='Switch sides, become observer, or release faction'):
        '''Add a player roster to the module
        
        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        buttonText : str
            Text on button
        buttonTooltip : str
            Tool tip to show when hovering over button
        '''
        super(PlayerRoster,self).__init__(doc,self.TAG,node=node,
                                          buttonKeyStroke = buttonKeyStroke,
                                          buttonText      = buttonText,
                                          buttonToolTip   = buttonToolTip)
    def addSide(self,name):
        '''Add a `Side` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Side
            The added element
        '''
        return self.add(PlayerSide,name=name)
    def getSides(self):
        '''Get all sides'''
        return self.getAllElements(PlayerSide,False)
    def encode(self):
        '''Encode for save'''
        return ['PLAYER\ta\ta\t<observer>']

registerElement(PlayerRoster)

# --------------------------------------------------------------------
class PlayerSide(Element):
    TAG = 'entry'
    UNIQUE = ['name']
    def __init__(self,doc,node=None,name=''):
        '''Adds a side to the player roster

        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        name : str
            Name of side 
        '''
        super(PlayerSide,self).__init__(doc,self.TAG,node=node)
        if node is None:
            self.addText(name)

    def getPlayerRoster(self):
        '''Get Parent element'''
        return self.getParent(PlayerRoster)

registerElement(PlayerSide)


#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from .. common import Verbose
from . features import Features
## END_IMPORT



class BaseArchive:
    WORD_SIZE = 2
    
    def __init__(self,filename,mode='rb'):
        '''Read data from a MFT CArchive stored on disk

        Works as a context manager 
        '''
        with VerboseGuard(f'Opening archive {filename}'):
            self._filename = filename
            self._file = open(filename,mode)
            self._i    = 0
            self.vmsg  = lambda *args : Verbose().message(*args)
            #self.vmsg  = lambda *args : None
        
    def __enter__(self):
        '''Enter context'''
        return self

    def __exit__(self,*args,**kwargs):
        '''Exit context'''
        self._file.close()

    def tell(self):
        pass

    def read(self,n):
        '''Read n bytes from archive'''
        pass

    def chr(self,n):
        '''Read n characters from archive'''
        b = self.read(n)
        try:
            c = b.decode()
            self.vmsg(f'char->{c}')
            return c
        except:
            print(f'Failed at {b} ({self._file.tell()})')
            raise

    def int(self,n):
        '''Read an (unsigned) integer from archive'''
        b = self.read(n)
        i = int.from_bytes(b,'little' if Features().little_endian else 'big')
        self.vmsg(f'int->{i}')
        return i     

    def byte(self):
        '''Read a byte from archive'''
        return self.int(1)

    def word(self):
        '''Read a word (16bit integer) from archive'''
        w = self.int(BaseArchive.WORD_SIZE)
        self.vmsg(f'word->{w}')
        return w;


    def dword(self):
        '''Read a double word (32bit integer) from archive'''
        d = self.int(2*BaseArchive.WORD_SIZE)
        self.vmsg(f'dword->{d}')
        return d

    def size(self):
        '''Read a size'''
        s = self.int(Features().size_size)
        self.vmsg(f'size->{s}')
        return s

    def sub_size(self):
        '''Read a size'''
        s = self.int(Features().sub_size)
        self.vmsg(f'sub->{s}')
        return s
    
    def count(self):
        '''Read a count'''
        if Features().size_size == 4:
            return self.word()

        c = self.word()
        if c != 0xFFFF:
            return c

        c = self.dword()
        if c != 0xFFFFFFFF:
            return c

        return int(8)
        
    def iden(self):
        '''Read an identifier'''
        i = self.int(Features().id_size)
        self.vmsg(f'id->{i}')
        return i
        
    def _strlen(self):
        '''Read length of following string from archive'''
        # See https://github.com/pixelspark/corespark/blob/master/Libraries/atlmfc/src/mfc/arccore.cpp

        s = 1
        l = self.byte()
        
        if l < 0xFF: # Small ASCII string
            self.vmsg(f'slen->{l},{s}')
            return l, s
        
        l = self.word()
        if l == 0xFFFE: # Unicode  - try again
            s = 2
            l = self.byte()

            if l < 0xFF: # Small unicode
                self.vmsg(f'slen->{l},{s}')
                return l, s
            

            l = self.word() # Large unicode string

        if l < 0xFFFF: # Not super long
            self.vmsg(f'slen->{l},{s}')
            return l, s

        l = self.dword()

        if l < 0xFFFFFFFF: # Not hyper long
            self.vmsg(f'slen->{l},{s}')
            return l, s

        
        self.vmsg(f'slen->{8},fixed')
        return self.int(8)
        
    def str(self):
        '''Read a string from the archive'''
        # See https://github.com/pixelspark/corespark/blob/master/Libraries/atlmfc/src/mfc/arccore.cpp
        l, s = self._strlen()
        # print(f'Read string of length {l}*{s} at {self.tell()}')
        ret  = [self.read(s) for i in range(l)]
        try:
            ret = [c.decode() for c in ret]
        except:
            ret = ['']
        self.vmsg(f'str->"{"".join(ret)}"')
        return ''.join(ret)

    @property
    def filename(self):
        return self._filename

    @property
    def path(self):
        from pathlib import Path
        return Path(self._filename)
    
# ====================================================================
class UnbufferedArchive(BaseArchive):
    def __init__(self,filename,mode='rb'):
        '''Read data from a MFT CArchive stored on dist

        Works as a context manager 
        '''
        super(UnbufferedArchive,self).__init__(filename,mode='rb')

    def read(self,n):
        '''Read n bytes from archive - directly from file'''
        b = self._file.read(n)
        self.vmsg(f'read->{list(b)}')
        # print(f'{self._i:6d} -> {b}')
        self._i += n
        return b

    def tell(self):
        return self._file.tell()

# ====================================================================
class BufferedArchive(BaseArchive):
    def __init__(self,filename,mode='rb'):
        '''Read data from a MFT CArchive stored on dist

        Works as a context manager 
        '''
        super(BufferedArchive,self).__init__(filename,mode='rb')
        self._size    = 4096
        self._max     = self._size
        self._current = self._max
        self._buffer  = []

    def read(self,n):
        with VerboseGuard(f'Read {n} bytes') as g:
            '''Read n bytes from archive - buffered
            
            This emulates the behaviour of MFC CArchive::Read 
            '''
            
            nmax          = n
            ntmp          =  min(nmax, self._max - self._current)
            b             =  self._buffer[self._current:self._current+ntmp]
            g(f'Take {ntmp} bytes from buffer -> {b}')
            self._current += ntmp
            nmax          -= ntmp
            
            if nmax != 0:
                g(f'Need to read at least {nmax} from file')
                assert self._current == self._max,\
                    f'Something is wrong! {self._current} != ' \
                    f'{self._max} (1)'

                g(f'Missing {nmax} bytes -> ({nmax % self._size})')
                ntmp         = nmax - (nmax % self._size)
                nread        = 0
                nleft        = ntmp
                nbytes       = 0
                while True:
                    tmpbuf =  self._file.read(nleft)
                    nbytes =  len(tmpbuf)
                    nread  += nbytes
                    nleft  -= nbytes
                    b      += tmpbuf
                    g(f'Read {nleft} -> {tmpbuf}')
                    
                    if nbytes <= 0 or nleft <= 0:
                        break
            
                nmax -= nread
            
                if nmax > 0 and nread  == ntmp:
                    # Last read chunk into buffer and copy
                    assert self._current == self._max,\
                        f'Something is wrong! {self._current} != ' \
                        f'{self._max} (2)'
                    
                    assert nmax < self._size, \
                        f'Something is wrong {nmax} >= {self._size}'
                    
                    nlastleft    = max(nmax,self._size)
                    nlastbytes   = 0
                    nread        = 0
                    self._buffer = []
                    while True:
                        tmpbuf       =  self._file.read(nlastleft)
                        nlastbytes   =  len(tmpbuf)
                        nread        += nlastbytes
                        nlastleft    -= nlastbytes
                        self._buffer += tmpbuf
            
                        if (nlastbytes <= 0) or \
                           (nlastleft <= 0) or \
                           nread >= ntmp:
                            break
            
                    self._current = 0
                    self._max     = nread
            
                    ntmp          =  min(nmax, self._max - self._current)
                    b             += self._buffer[self._current:
                                                  self._current+ntmp]
                    self._current += ntmp
                    nmax          -= ntmp
            
            g(b)
            return b''.join(b)

    def tell(self):
        return self._file.tell()
    

Archive = UnbufferedArchive
# Archive = BufferedArchive

#
# EOF
#



#!/usr/bin/env python
## BEGIN_IMPORTS
from pywargame.zuntzu.gamebox  import ZTGameBox
from pywargame.zuntzu.exporter import ZTExporter
from pywargame.common          import Verbose
from pywargame.zuntzu.base     import ZTImage
## END_IMPORTS

# ====================================================================
def exportMain():
    from argparse import ArgumentParser, FileType, \
         RawDescriptionHelpFormatter 
    from pathlib import Path
    from textwrap import wrap, dedent

    w = lambda s : '\n\n'.join(['\n'.join(wrap(p))
                                for p in dedent(s).split('\n\n')])
    e = '''
           If the script fails with "cache resources exhausted",
           try reducing the resolution (-r).

           If the ZunTzu game box does not have selectable
           boards, use the -a option to add all boards to
           the generated VASSAL module.

           Pass a Python script with the option -p to do
           user post-processing with that script.  Use the
           pywargame VASSAL API to do all sorts of manipulations.
           '''
    
    ap = ArgumentParser(description='Read in a ZunTzu GameBox',
                        formatter_class= RawDescriptionHelpFormatter,
                        epilog=w(e))
    ap.add_argument('input', type=FileType('rb'), 
                    help='Input ZunTzu GameBox')
    ap.add_argument('--output', '-o', type=str, nargs='?',
                    default='', help='Output VASSAL module')
    ap.add_argument('-p','--patch',
                    help='A python script to patch generated module',
                    type=FileType('r'))
    ap.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')
    ap.add_argument('-r','--resolution',help='Set target DPI',type=int,
                    choices=[75,150,300,600],default=150)
    ap.add_argument('-a','--all-maps',action='store_true',
                    help='Generate all maps')
    ap.add_argument('-1','--one-map',action='store_false',dest='all_maps',
                    help='All maps are different versions of main map')
    ap.add_argument('-R','--n-rotations',type=int,
                    help='Number of rotations to allow. Set to 1 for '
                    'arbitrary rotations',default=12)
    ap.add_argument('-W','--vassal-version',
                    help='Vassal version number',
                    type=str,
                    default='3.7.0')
    ap.add_argument('-S','--symbolic-dice',
                    help='Use symbolic dice',
                    action='store_true')
    ap.add_argument('-T','--text-dice',
                    help='Use normal text dice',
                    dest='symbolic_dice',
                    action='store_false')
    ap.add_argument('-v','--version',
                    help='Override version',
                    type=str,
                    default='0.0.1')
    ap.set_defaults(symbolic_dice=True,
                    all_maps=False)


    args = ap.parse_args()

    Verbose().setVerbose(args.verbose)

    ZTImage.target_dpi = args.resolution
    output = args.output
    patch  = args.patch 
    if output == '':
        p      = Path(args.input.name)
        output = p.with_suffix('.vmod')
    if patch is not None:
        tmp = patch.name
        patch.close()
        patch = tmp
        
    gb = ZTGameBox(args.input)
    # gb.write_images()
    
    try:
        exporter = ZTExporter(gb,
                              version=args.version,
                              vassalVersion=args.vassal_version,
                              allmaps=args.all_maps,
                              nrotations=args.n_rotations,
                              symbolic_dice=args.symbolic_dice)
        exporter.run(output,patch)
    except Exception as e:
        from sys import stderr 
        from os import unlink
        
        print(f'Failed to build {output}: {e}',file=stderr)

        try:
            unlink(vmodname)
        except:
            pass

        raise e
    
# ====================================================================
if __name__ == '__main__':
    exportMain()
        
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT


class MoveFixedTrait(Trait):
    ID = 'translate'

    def __init__(self,
                 command        = '',   # menu command,
                 key            = '',   # Hotkey or command
                 dx             = 0,    # X distance (int or expr))
                 dy             = 0,    # Y distance (int or expr))
                 stack          = False,# Move entire stack
                 xStepFactor    = 0,    # Factor on X offset (int or expr)
                 yStepFactor    = 0,    # Factor on Y offset (int or expr)
                 xStep          = 0,    # X offset (int or expr)
                 yStep          = 0,    # Y offset (int or expr)
                 description    = '',   # str 
                 ):
        '''Move a fixed distance.

           x' = dx + xStepFactor * xStep 
           y' = dy + yStepFactor * yStep 

        If piece can rotate, and this is trait is given _after_ (more
        recent in the traits list), then the piece will move according
        to the direction faced. If given _before_ (later in the traits list),
        then the move is relative  to the current map. 
        
        (VASSAL.counters.Translate)

        '''
        self.setType(command     = command,
                     key         = key,
                     dx          = dx,
                     dy          = dy,
                     stack       = stack,
                     xStepFactor = xStepFactor,
                     yStepFactor = yStepFactor,
                     xStep       = xStep,
                     yStep       = yStep,
                     description = description)
        self.setState()


Trait.known_traits.append(MoveFixedTrait)
#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . image import GBXImage
from . draw import GBXDraw
from . features import Features
## END_IMPORT

# ====================================================================
class GBXCellGeometry:
    RECTANGLE        = 0
    HORIZONTAL_BRICK = 1
    VERTICAL_BRICK   = 2
    HEXAGON          = 3
    SIDEWAYS_HEXAGON = 4
    STAGGER_OUT      = 0
    STAGGER_IN       = 1
    TYPES = {
        RECTANGLE       : 'rectangle',
        HORIZONTAL_BRICK: 'horizontal brick',
        VERTICAL_BRICK  : 'vertical brick',
        HEXAGON         : 'hexagon',
        SIDEWAYS_HEXAGON: 'sideways hexagon'
    }
    STAGGERS = {
        STAGGER_OUT: 'out',
        STAGGER_IN:  'in'
    }

    def __init__(self,ar):
        '''The geometry of cells'''
        with VerboseGuard('Reading cell geometry'):
            from numpy import max
            
            self._type      = ar.word()
            self._stagger   = ar.word()
            self._left      = ar.word()
            self._top       = ar.word()
            self._right     = ar.word()
            self._bottom    = ar.word()
            n               = 7 if self._type > 2 else 5
            self._points    = [[ar.word(),ar.word()] for _ in range(n)]
            size            = max(self._points,axis=0)
            self._dx        = int(size[0])
            self._dy        = int(size[1])
            self._size      = [self._dx,self._dy]
            
            if self._type == self.HEXAGON:
                self._dx = int(0.75 * self._dx)
            elif self._type == self.SIDEWAYS_HEXAGON:
                self._dy = int(0.75 * self._dy)

    def toDict(self):
        from numpy import max
        return {'shape':      self.TYPES.get(self._type,''),
                'stagger':    self.STAGGERS.get(self._stagger,''),
                'size':       self._size,
                'bounding box (ltrb)':
                (self._left, self._top, self._right, self._bottom),
                'points':     self._points }

    def svgDef(self,dwg):
        with VerboseGuard('Defining SVG cell geometry'):
            if self._type in [0,1,2]:
                return dwg.defs.add(dwg.rect(id='cell',
                                             size=(self._right-self._left,
                                                   self._bottom-self._top)))
        
            return dwg.defs.add(dwg.polygon(self._points,id='cell'))

    def translate(self,row,col,center=False):
        x = col * self._dx
        y = row * self._dy
        if self._type == self.RECTANGLE: # No offset for rectangles
            return x,y
        if self._type in [self.HORIZONTAL_BRICK,self.SIDEWAYS_HEXAGON]:
            x += self._dx//2 if (row % 2) != self._stagger else 0
        if self._type in [self.VERTICAL_BRICK,self.HEXAGON]:
            y += self._dy//2 if (col % 2) != self._stagger else 0
        if center:
            x += self._size[0]//2
            y += self._size[1]//2
        return x,y

    def inverse(self,x,y):
        col = x / self._dx
        row = y / self._dy
        if self._type in [self.HORIZONTAL_BRICK,self.SIDEWAYS_HEXAGON]:
            col -= .5 if (int(row) % 2) != self._stagger else 0
        if self._type in [self.VERTICAL_BRICK,self.HEXAGON]:
            row -= .5  if (int(col) % 2) != self._stagger else 0

        # CyberBoard start at 1
        return int(row)+1, int(col)+1 
        

    def boardSize(self,nrows,ncols):
        w = ncols * self._dx
        h = nrows * self._dy

        if self._type in [2,3]:
            h += self._dy // 2
        if self._type in [1,4]:
            w += self._dx // 2
        if self._type == 3:
            w += self._dx // 3
        if self._type == 4:
            h += self._dy // 3

        return w+1,h+1
    
    def __str__(self):
        return (f'type: {self.TYPES.get(self._type,"")} '
                + f'stagger: {self.STAGGERS.get(self._stagger,"")} '
                + f'({self._left},{self._top})x({self._right},{self._bottom}) '
                + f': [{self._points}]')

        
# --------------------------------------------------------------------
class GBXCell:
    def __init__(self,ar,row,column):
        '''A single cell'''
        with VerboseGuard(f'Reading cell row={row} column={column}'):
            self._row    = row
            self._column = column
            if Features().id_size == 4:
                self._is_tile = ar.byte();
            self._tile   = ar.dword()
            if Features().id_size != 4:
                self._is_tile  = (self._tile >> 16) == 0xFFFF;
                if self._is_tile:
                    self._tile = self._tile & 0xFFFF

    def tileID(self):
        if not self._is_tile:
            return None
        return self._tile

    def color(self):
        if self._is_tile:
            return None
        return GBXDraw.hex(self._tile)

    def toDict(self,tileManager,calc=None):
        d = {'row': self._row,
             'column': self._column}
        if not self._is_tile:
            d['color'] = GBXDraw.hex(self._tile)
        else:
            d['tile'] = tileManager.store(self._tile)
        if calc is not None:
            d['pixel'] = calc.translate(self._row,self._column,True)
        return d
    
    
    def svgDef(self,dwg,tileManager,ptm):
        tileID = self.tileID()
        if tileID is None:
            return

        if tileID in ptm: # Have it
            return 

        with VerboseGuard(f'Defining SVG pattern'):
            img  = tileManager.image(tileID)
            data = GBXImage.b64encode(img)
            if data is None:
                return
            
            iden = f'terrain_{tileID:04x}'
            pat  = dwg.defs.add(dwg.pattern(id=iden,
                                            size=(img.width,img.height)))
            pat.add(dwg.image(href=(data)))
            ptm[tileID] = pat

    def svg(self,dwg,g,cell,geom,ptm):
        tileID = self.tileID()
        if tileID is not None:
            fill = ptm[tileID].get_paint_server()
        else:
            fill = self.color()

        trans = geom.translate(self._row,self._column)
        iden  = f'cell_bg_{self._column:03d}{self._row:03d}'
        g.add(dwg.use(cell,insert=trans,fill=fill,id=iden))
        
    def svgFrame(self,dwg,g,cell,geom,color):
        trans = geom.translate(self._row,self._column)
        iden  = f'cell_fg_{self._column:03d}{self._row:03d}'
        g.add(dwg.use(cell,
                      insert=trans,
                      stroke=GBXDraw.hex(color),
                      fill='none',
                      id=iden))
        

    def __str__(self):
        return f'({self._row:02d},{self._column:02d}): {self._tile:08x}'

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . gameelements import GameElement
from . widget import WidgetElement
## END_IMPORT

# --------------------------------------------------------------------
class ChartWindow(GameElement,WidgetElement):
    TAG=Element.MODULE+'ChartWindow'
    UNIQUE = ['name']
    def __init__(self,elem,node=None,
                 name        = '',
                 hotkey      = key('A',ALT),
                 description = '',
                 text        = '',
                 tooltip     = 'Show/hide Charts',
                 icon        = '/images/chart.gif'):
        super(ChartWindow,self).__init__(elem,self.TAG,node=node,
                                         name        = name,
                                         hotkey      = hotkey,
                                         description = description,
                                         text        = text,
                                         tooltip     = tooltip,
                                         icon        = icon)

    def addChart(self,**kwargs):
        '''Add a `Chart` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Chart
            The added element
        '''
        return self.add(Chart,**kwargs)
    def getCharts(self,asdict=True):
        '''Get all Chart element(s) from this

        Parameters
        ----------
        asdict : bool        
            If `True`, return a dictonary that maps key to `Chart`
            elements.  If `False`, return a list of all Chart`
            children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Chart` children

        '''
        return self.getElementsById(Chart,'chartName',asdict=asdict)
    
registerElement(ChartWindow)    

# --------------------------------------------------------------------
class Chart(Element):
    TAG=Element.WIDGET+'Chart'
    UNIQUE = ['chartName','fileName']
    def __init__(self,elem,node=None,
                 chartName   = '',
                 fileName    = '',
                 description = ''):
        super(Chart,self).__init__(elem,self.TAG,node=node,
                                   chartName   = chartName,
                                   description = description,
                                   fileName    = fileName)

registerElement(Chart)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . documentation import Documentation
from . gameelements import *
from . widget import *
from . player import *
from . globalproperty import *
from . chessclock import *
from . turn import TurnTrack
from . map import Map
from . mapelements import *
from . chart import *
from . board import *
## END_IMPORT

# --------------------------------------------------------------------
class Game(Element):
    TAG = Element.BUILD+'GameModule'
    UNIQUE = ['name']
    def __init__(self,build,node=None,
                 name            = '',
                 version         = '', 
                 ModuleOther1    = "",
                 ModuleOther2    = "",
                 VassalVersion   = "3.6.7",
                 description     = "",
                 nextPieceSlotId = 20):
        '''Create a new Game object

        Parameters
        ----------
        build : xml.dom.Document
            root note
        node : xml.dom.Node
            To read from, or None
        name : str
            Name of module
        version : str
            Version of module
        ModuleOther1 : str
            Free form string 
        ModuleOther2 : str
            Free form string
        VassalVersion : str
            VASSAL version this was created for
        description : str
            Speaks volumes
        nextPieceSlotId : int
            Starting slot ID.
        '''
        super(Game,self).__init__(build, self.TAG,
                                  node            = node,
                                  name            = name,
                                  version         = version,
                                  ModuleOther1    = ModuleOther1,
                                  ModuleOther2    = ModuleOther2,
                                  VassalVersion   = VassalVersion,
                                  description     = description,
                                  nextPieceSlotId = nextPieceSlotId)
    def nextPieceSlotId(self):
        '''Increment next piece slot ID'''
        ret = int(self.getAttribute('nextPieceSlotId'))
        self.setAttribute('nextPieceSlotId',str(ret+1))
        return ret
    #
    def addBasicCommandEncoder(self,**kwargs):
        '''Add a `BasicCommandEncoder` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : BasicCommandEncoder
            The added element
        '''
        return self.add(BasicCommandEncoder,**kwargs)
    def addGlobalTranslatableMessages(self,**kwargs):
        '''Add a `GlobalTranslatableMessages` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : GlobalTranslatableMessages
            The added element
        '''
        return self.add(GlobalTranslatableMessages,**kwargs)
    def addPlayerRoster(self,**kwargs):
        '''Add a `PlayerRoster` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : PlayerRoster
            The added element
        '''
        return self.add(PlayerRoster,**kwargs)
    def addChessClock(self,**kwargs):
        '''Add a `ChessClockControl` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : PlayerRoster
            The added element
        '''
        return self.add(ChessClockControl,**kwargs)
    def addLanguage(self,**kwargs):
        '''Add a `Language` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Language
            The added element
        '''
        return self.add(Language,**kwargs)
    def addChatter(self,**kwargs):
        '''Add a `Chatter` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Chatter
            The added element
        '''
        return self.add(Chatter,**kwargs)
    def addKeyNamer(self,**kwargs):
        '''Add a `KeyNamer` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : KeyNamer
            The added element
        '''
        return self.add(KeyNamer,**kwargs)
    def addNotes(self,**kwargs):
        '''Add a `Notes` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Notes
            The added element
        '''
        return self.add(Notes,**kwargs)
    def addLanguage(self,**kwargs):
        '''Add a `Language` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Language
            The added element
        '''
        return self.add(Language,**kwargs)
    def addChatter(self,**kwargs):
        '''Add a `Chatter` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Chatter
            The added element
        '''
        return self.add(Chatter,**kwargs)
    def addKeyNamer(self,**kwargs):
        '''Add a `KeyNamer` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : KeyNamer
            The added element
        '''
        return self.add(KeyNamer,**kwargs)
    def addGlobalProperties(self,**kwargs):
        '''Add a `GlobalProperties` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : GlobalProperties
            The added element
        '''
        return self.add(GlobalProperties,**kwargs)
    def addGlobalOptions(self,**kwargs):
        '''Add a `GlobalOptions` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : GlobalOptions
            The added element
        '''
        return self.add(GlobalOptions,**kwargs)
    def addTurnTrack(self,**kwargs):
        '''Add a `TurnTrack` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : TurnTrack
            The added element
        '''
        return self.add(TurnTrack,**kwargs)
    def addDocumentation(self,**kwargs):
        '''Add a `Documentation` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Documentation
            The added element
        '''
        return self.add(Documentation,**kwargs)
    def addPrototypes(self,**kwargs):
        '''Add a `Prototypes` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Prototypes
            The added element
        '''
        return self.add(Prototypes,**kwargs)
    def addPieceWindow(self,**kwargs):
        '''Add a `PieceWindow` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : PieceWindow
            The added element
        '''
        return self.add(PieceWindow,**kwargs)
    def addChartWindow(self,**kwargs):
        '''Add a `ChartWindow` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : ChartWindow
            The added element
        '''
        return self.add(ChartWindow,**kwargs)
    def addInventory(self,**kwargs):
        '''Add a `Inventory` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Inventory
            The added element
        '''
        return self.add(Inventory,**kwargs)
    def addMap(self,**kwargs):
        '''Add a `Map` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Map
            The added element
        '''
        return self.add(Map,**kwargs)
    def addDiceButton(self,**kwargs):
        '''Add a `DiceButton` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : DiceButton
            The added element
        '''
        return self.add(DiceButton,**kwargs)
    def addPredefinedSetup(self,**kwargs):
        '''Add a `PredefinedSetup` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : PredefinedSetup
            The added element
        '''
        return self.add(PredefinedSetup,**kwargs)
    def addGameMassKey(self,**kwargs):
        '''Add a `GameMassKey` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : GameMassKey
            The added element
        '''
        return self.add(GameMassKey,**kwargs)
    def addStartupMassKey(self,**kwargs):
        '''Add a `StartupMassKey` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : StartupMassKey
            The added element
        '''
        return self.add(StartupMassKey,**kwargs)
    def addMenu(self,**kwargs):
        '''Add a `Menu` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Menu
            The added element
        '''
        return self.add(Menu,**kwargs)
    def addSymbolicDice(self,**kwargs):
        '''Add a `SymbolicDice` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : SymbolicDice
            The added element
        '''
        return self.add(SymbolicDice,**kwargs)

    def addFolder(self,**kwargs):
        '''Add a `ModuleFolder` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : ModuleFolder
            The added element
        '''
        return self.add(ModuleFolder,**kwargs)
    
    
    # ----------------------------------------------------------------
    def getGlobalProperties(self,single=True):
        '''Get all or a sole `GlobalPropertie` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `GlobalPropertie` child, otherwise fail.
            If `False` return all `GlobalPropertie` children in this element
        
        Returns
        -------
        children : list
            List of `GlobalPropertie` children (even if `single=True`)
        '''
        return self.getAllElements(GlobalProperties,single)
    def getBasicCommandEncoder(self,single=True):
        '''Get all or a sole `Ba` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Ba` child, otherwise fail.
            If `False` return all `Ba` children in this element
        
        Returns
        -------
        children : list
            List of `Ba` children (even if `single=True`)
        '''
        return self.getAllElements(BasicCommandEncoder,single)
    def getGlobalTranslatableMessages(self,single=True):
        '''Get all or a sole `GlobalTranslatableMessage` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `GlobalTranslatableMessage` child, otherwise fail.
            If `False` return all `GlobalTranslatableMessage` children in this element
        
        Returns
        -------
        children : list
            List of `GlobalTranslatableMessage` children (even if `single=True`)
        '''
        return self.getAllElements(GlobalTranslatableMessages,single)
    def getLanguages(self,single=False):
        '''Get all or a sole `Language` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Language` child, otherwise fail.
            If `False` return all `Language` children in this element
        
        Returns
        -------
        children : list
            List of `Language` children (even if `single=True`)
        '''
        return self.getAllElements(Language,single)
    def getChessClocks(self,asdict=False):
        '''Get all or a sole `Language` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Language` child, otherwise fail.
            If `False` return all `Language` children in this element
        
        Returns
        -------
        children : list
            List of `Language` children (even if `single=True`)
        '''
        return self.getElementsByKey(ChessClockControl,'name',asdict)
    def getChatter(self,single=True):
        '''Get all or a sole `Chatter` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Chatter` child, otherwise fail.
            If `False` return all `Chatter` children in this element
        
        Returns
        -------
        children : list
            List of `Chatter` children (even if `single=True`)
        '''
        return self.getAllElements(Chatter,single)
    def getKeyNamer(self,single=True):
        '''Get all or a sole `KeyNamer` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `KeyNamer` child, otherwise fail.
            If `False` return all `KeyNamer` children in this element
        
        Returns
        -------
        children : list
            List of `KeyNamer` children (even if `single=True`)
        '''
        return self.getAllElements(KeyNamer,single)
    def getDocumentation(self,single=True):
        '''Get all or a sole `Documentation` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Documentation` child, otherwise fail.
            If `False` return all `Documentation` children in this element
        
        Returns
        -------
        children : list
            List of `Documentation` children (even if `single=True`)
        '''
        return self.getAllElements(Documentation,single)
    def getPrototypes(self,single=True):
        '''Get all or a sole `Prototypes` (i.e., the containers of
        prototypes, not a list of actual prototypes) element(s) from
        this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Prototypes` child, otherwise fail.
            If `False` return all `Prototypes` children in this element
        
        Returns
        -------
        children : list
            List of `Prototype` children (even if `single=True`)

        '''
        return self.getAllElements(Prototypes,single)
    def getPlayerRoster(self,single=True):
        '''Get all or a sole `PlayerRo` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `PlayerRo` child, otherwise fail.
            If `False` return all `PlayerRo` children in this element
        
        Returns
        -------
        children : list
            List of `PlayerRo` children (even if `single=True`)
        '''
        return self.getAllElements(PlayerRoster,single)
    def getGlobalOptions(self,single=True):
        '''Get all or a sole `GlobalOption` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `GlobalOption` child, otherwise fail.
            If `False` return all `GlobalOption` children in this element
        
        Returns
        -------
        children : list
            List of `GlobalOption` children (even if `single=True`)
        '''
        return self.getAllElements(GlobalOptions,single)
    def getInventories(self,asdict=True):
        '''Get all Inventorie element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Inventorie` elements.  If `False`, return a list of all Inventorie` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Inventorie` children
        '''
        return self.getElementsByKey(Inventory,'name',asdict)
    def getPieceWindows(self,asdict=True):
        '''Get all PieceWindow element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `PieceWindow` elements.  If `False`, return a list of all PieceWindow` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `PieceWindow` children
        '''
        return self.getElementsByKey(PieceWindow,'name',asdict)
    def getChartWindows(self,asdict=True):
        '''Get all ChartWindow element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `ChartWindow` elements.  If `False`, return a list of all ChartWindow` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `ChartWindow` children
        '''
        return self.getElementsByKey(ChartWindow,'name',asdict)
    def getDiceButtons(self,asdict=True):
        '''Get all DiceButton element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `DiceButton` elements.  If `False`, return a list of all DiceButton` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `DiceButton` children
        '''
        return self.getElementsByKey(DiceButton,'name',asdict)
    def getPredefinedSetups(self,asdict=True):
        '''Get all PredefinedSetup element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `PredefinedSetup` elements.  If `False`, return a list of all PredefinedSetup` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `PredefinedSetup` children
        '''
        return self.getElementsByKey(PredefinedSetup,'name',asdict)
    def getNotes(self,single=True):
        '''Get all or a sole `Note` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Note` child, otherwise fail.
            If `False` return all `Note` children in this element
        
        Returns
        -------
        children : list
            List of `Note` children (even if `single=True`)
        '''
        return self.getAllElements(Notes,single)
    def getTurnTracks(self,asdict=True):
        '''Get all TurnTrack element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `TurnTrack` elements.  If `False`, return a list of all TurnTrack` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `TurnTrack` children
        '''
        return self.getElementsByKey(TurnTrack,'name',asdict)
    def getPieces(self,asdict=False):
        '''Get all Piece element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Piece` elements.  If `False`, return a list of all Piece` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Piece` children
        '''
        return self.getElementsByKey(PieceSlot,'entryName',asdict)
    def getCards(self,asdict=False):
        '''Get all Cards element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Piece` elements.  If `False`, return a list of all Piece` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Piece` children
        '''
        return self.getElementsByKey(CardSlot,'entryName',asdict)
    def getSpecificPieces(self,*names,asdict=False):
        '''Get all SpecificPiece element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `SpecificPiece` elements.  If `False`, return a list of all SpecificPiece` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `SpecificPiece` children
        '''
        return self.getSpecificElements(PieceSlot,'entryName',
                                        *names,asdict=asdict)
    def getMap(self,asdict=False):
        return self.getElementsByKey(Map,'mapName',asdict)
    def getWidgetMaps(self,asdict=True):
        '''Get all WidgetMap element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `WidgetMap` elements.  If `False`, return a list of all WidgetMap` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `WidgetMap` children
        '''
        return self.getElementsByKey(WidgetMap,'mapName',asdict=asdict)
    def getMaps(self,asdict=True):
        '''Get all Map element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Map` elements.  If `False`, return a list of all Map` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Map` children
        '''
        maps = self.getMap(asdict=asdict)
        wmaps = self.getWidgetMaps(asdict=asdict)
        if asdict:
            maps.update(wmaps)
        else:
            maps.extend(wmaps)
        return maps
    def getBoards(self,asdict=True):
        '''Get all Board element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Board`
            elements.  If `False`, return a list of all Board`
            children.

        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Board` children

        '''
        return self.getElementsByKey(Board,'name',asdict)
    def getGameMassKeys(self,asdict=True):
        '''Get all GameMassKey element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Board`
            elements.  If `False`, return a list of all Board`
            children.

        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Board` children

        '''
        return self.getElementsByKey(GameMassKey,'name',asdict)
    def getStartupMassKeys(self,asdict=True):
        '''Get all StartupMassKey element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Board`
            elements.  If `False`, return a list of all Board`
            children.

        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Board` children

        '''
        return self.getElementsByKey(StartupMassKey,'name',asdict)
    def getMenus(self,asdict=True):
        '''Get all Menu element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Board`
            elements.  If `False`, return a list of all Board`
            children.

        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Board` children

        '''
        return self.getElementsByKey(Menu,'text',asdict)
    def getSymbolicDices(self,asdict=True):
        '''Get all Menu element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `SymbolicDice`
            elements.  If `False`, return a list of all `SymbolicDice`
            children.

        Returns
        -------
        children : dict or list
            Dictionary or list of `SymbolicDice` children

        '''
        return self.getElementsByKey(SymbolicDice,'name',asdict)
    def getFolders(self,asdict=True):
        '''Get all Menu element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Folder`
            elements.  If `False`, return a list of all `Folder`
            children.

        Returns
        -------
        children : dict or list
            Dictionary or list of `Folder` children

        '''
        return self.getElementsByKey(ModuleFolder,'name',asdict)
    
    def getAtStarts(self,single=False):
        '''Get all or a sole `AtStart` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `AtStart` child, otherwise fail.
            If `False` return all `AtStart` children in this element
        
        Returns
        -------
        children : list
            List of `AtStart` children (even if `single=True`)
        '''
        return self.getAllElements(AtStart,single)

registerElement(Game)

# --------------------------------------------------------------------
# Old game module class
class OldGame(Game):
    TAG = 'VASSAL.launch.BasicModule'
    def __init__(self,build,**kwargs):
        super().__init__(build,**kwargs)
    
registerElement(OldGame)

# --------------------------------------------------------------------
class BasicCommandEncoder(GameElement):
    TAG = Element.MODULE+'BasicCommandEncoder'
    def __init__(self,doc,node=None):
        super(BasicCommandEncoder,self).__init__(doc,self.TAG,node=node)

registerElement(BasicCommandEncoder)

#
# EOF
#



## BEGIN_IMPORT
from . save import SaveData, SaveFile, SaveIO
from . moduledata import ModuleData
from . vmod import VMod
## END_IMPORT

# --------------------------------------------------------------------
class VSav:
    SAVE_DATA = 'savedata'
    
    def __init__(self,build,vmod):
        '''Create a VASSAL save file programmatically

        Parameters
        ----------
        build : xml.dom.Document
            `buildFile.xml` as XML
        vmod : VMod
            Module file
        '''
        from time import time 
        self._vmod  = vmod
        self._game  = build.getGame()
        self._start = int(time()*1000)
        

    def createSaveData(self,description=None):
        '''Create `savedgame`'''
        desc           = (self._game['description']
                          if description is None else description)
        self._saveData = SaveData(root=None)
        data           = self._saveData.addData()
        data.addVersion      (version    =self._game['version'])
        data.addVASSALVersion(version    =self._game['VassalVersion'])
        data.addDescription  (description=desc)
        data.addDateSaved    (milisecondsSinceEpoch=self._start)
        return self._saveData

    def createModuleData(self):
        '''Create `moduleData`'''
        self._moduleData = ModuleData()
        data = self._moduleData.addData()
        data.addVersion      (version    =self._game['version'])
        data.addVASSALVersion(version    =self._game['VassalVersion'])
        data.addName         (name       =self._game['name'])
        data.addDescription  (description=self._game['description'])
        data.addDateSaved    (milisecondsSinceEpoch=self._start)
        return self._moduleData
        
    def addSaveFile(self):
        '''Add a save file to the module

        Returns
        -------
        vsav : SaveFile
            Save file to add content to        
        '''
        self._saveFile = SaveFile(game=self._game,firstid=self._start)
        return self._saveFile

    def run(self,
            savename    = 'Save.vsav',
            description = None,
            update      = None):
        '''Run this to generate the save file

        Parameters
        ----------
        savename : str
            Name of save file to write
        description : str
            Short description of the save file
        update : callable or None
            A callable that can update trait states after the piece
            traits have been fully flattened.  The callable should
            adhere to the interface

                update(name,traits)

            where `name` is the name of the piece (entryName) and
            `traits` is a list of unrolled traits.

        '''
        from zipfile import ZipFile, ZIP_DEFLATED
        
        self.createSaveData(description=description)
        self.createModuleData()
        
        with self._vmod.getInternalFile(savename,'w') as vsav:
            with ZipFile(vsav,'w',ZIP_DEFLATED) as zvsav:
                # The key is set to 0xAA (alternating ones and zeros)
                SaveIO.writeInZip(zvsav,0xAA,
                                  self._saveFile.getLines(update=update))
            
                zvsav.writestr(VMod.MODULE_DATA, self._moduleData.encode())
                zvsav.writestr(VSav.SAVE_DATA,   self._saveData.encode())
            
#
# EOF
#



# ====================================================================
class Trait:
    known_traits = []
    def __init__(self):
        '''Base class for trait capture.
        
        Unlike the Element classes, this actually holds state that
        isn't reflected elsewhere in the DOM.  This means that the
        data here is local to the object.  So when we do::

        
            piece  = foo.getPieceSlots()[0]
            traits = p.getTraits()
            for trait in traits:
                if trait.ID == 'piece': 
                    trait["gpid"] = newPid
                    trait["lpid"] = newPid
                
        we do not actually change anything in the DOM.  To do that, we
        must add back _all_ the traits as::

            piece.setTraits(traits)
        
        We can add traits to a piece, like::

            piece.addTrait(MarkTrait('Hello','World'))
                
        But it is not particularly efficient.  Better to do
        (continuing from above)::

       
            traits.append(MarkTrait('Hello','World;)
            piece.setTraits(traits)
       
        
        .. include:: traits/README.md
           :parser: myst_parser.sphinx_
        
        '''
        self._type  = None
        self._state = None

    def setType(self,**kwargs):
        '''Set types.  Dictionary of names and values.  Dictonary keys
        defines how we access the fields, which is internal here.
        What matters is the order of the values.

        '''
        self._type   = list(kwargs.values())
        self._tnames = list(kwargs.keys())

    def setState(self,**kwargs):
        '''Set states.  Dictionary of names and values.  Dictonary keys
        defines how we access the fields, which is internal here.
        What matters is the order of the values.
        '''
        self._state  = list(kwargs.values())
        self._snames = list(kwargs.keys())

    def __getitem__(self,key):
        '''Look up item in either type or state'''
        try:
            return self._type[self._tnames.index(key)]
        except:
            pass
        return self._state[self._snames.index(key)]

    def __setitem__(self,key,value):
        '''Set item in either type or state'''
        try:
            self._type[self._tnames.index(key)] = value
            return
        except:
            pass
        self._state[self._snames.index(key)] = value

    def encode(self,term=False):
        '''
        returns type and state encoded'''
        t = self.encodeFields(self.ID,*self._type,term=term)
        s = self.encodeFields(*self._state,term=term)
        return t,s

    @classmethod
    def findTrait(cls,traits,ID,key=None,value=None,verbose=False):
        for trait in traits:
            if trait.ID != ID:
                continue
            if verbose:
                print(f' {trait.ID}')
            if key is None or value is None:
                if verbose:
                    print(f' Return {trait.ID}')
                return trait
            if verbose:
                print(f' Check {key}={value}: {trait[key]}')
            if trait[key] == value:
                return trait
        if verbose:
            print(f' Trait of type {ID} with {key}={value} not found')
        return None
        
    @classmethod
    def take(cls,iden,t,s):
        '''If the first part of the string t matches the ID, then take it.

        t and s are lists of strings.
        ''' 
        if iden != cls.ID: return None

        ret = cls()
        ret._type = t
        ret._state = s
        ret.check() # Check if we're reasonable, or raise
        #print(f'Took {iden} {cls}\n'
        #      f'  {ret._tnames}\n'
        #      f'  {ret._snames}')
        return ret

    def check(self):
        '''Implement if trait should check that all is OK when cloning'''
        pass

    @classmethod
    def encodeFields(cls,*args,term=False):
        return ';'.join([str(e).lower() if isinstance(e,bool) else str(e)
                         for e in args])+(';' if term else '')

    @classmethod
    def decodeFields(cls,s):
        from re import split
        return split(r'(?<!\\);',s)
        # return s.split(';') # Probably too simple-minded 

    @classmethod
    def encodeKeys(cls,keys,sep=','):
        return sep.join([k.replace(',','\\'+f'{sep}') for k in keys])
        
    @classmethod
    def decodeKeys(cls,keys,sep=','):
        from re import split
        ks = split(r'(?<!\\)'+f'{sep}',keys)
        return [k.replace('\\'+f'{sep}',f'{sep}') for k in ks]

    @classmethod
    def flatten(cls,traits,game=None,prototypes=None,verbose=False):
## BEGIN_IMPORT        
        from . traits import BasicTrait
## END_IMPORT        
        if prototypes is None:
            if game is None:
                print(f'Warning: Game or prototypes not passed')
                return None
            prototypes = game.getPrototypes()[0].getPrototypes()

        if len(traits) < 1: return None

        basic = None
        if traits[-1].ID == 'piece': # BasicTrait.ID:
            basic = traits.pop()

        if verbose:
            print(f'Piece {basic["name"]}')
            
        ret = cls._flatten(traits,prototypes,' ',verbose)
        ret.append(basic)

        return ret
    
    @classmethod
    def _flatten(cls,traits,prototypes,ind,verbose):
## BEGIN_IMPORT        
        from . traits import BasicTrait
        from . traits import PrototypeTrait
## END_IMPORT        
        '''Expand all prototype traits in traits'''
        ret = []
        for trait in traits:
            # Ignore recursive basic traits
            if trait.ID == BasicTrait.ID:
                continue
            # Add normal traits
            if trait.ID != PrototypeTrait.ID:
                if verbose:
                    print(f'{ind}Adding trait "{trait.ID}"')
                    
                ret.append(trait)
                continue

            # Find prototype
            name  = trait['name']
            proto = prototypes.get(name,None)
            if proto is None:
                if name != ' prototype':
                    print(f'{ind}Warning, prototype {name} not found')
                continue

            if verbose:
                print(f'{ind}Expanding prototype "{name}"')
                
            # Recursive call to add prototype traits (and possibly
            # more recursive calls 
            ret.extend(cls._flatten(proto.getTraits(), prototypes,
                                    ind+' ',verbose))

        return ret

    def print(self,file=None):
        if file is None:
            from sys import stdout
            file = stdout

        nt = max([len(i) for i in self._tnames]) if self._tnames else 0
        ns = max([len(i) for i in self._snames]) if self._snames else 0
        nw = max(nt,ns)

        print(f'Trait ID={self.ID}',file=file)
        print(f' Type:',            file=file)
        for n,v in zip(self._tnames,self._type):
            print(f'  {n:<{nw}s}: {v}',file=file)
        print(f' State:',           file=file)
        for n,v in zip(self._snames,self._state):
            print(f'  {n:<{nw}s}: {v}',file=file)
            
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class CloneTrait(Trait):
    ID      = 'clone'
    def __init__(self,
                 command         = '',
                 key             = '',
                 description     = ''):
        '''Create a clone trait (VASSAL.counter.Clone)'''
        super().__init__()

        self.setType(command         = command,          # Context menu name
                     key             = key,              # Context menu key
                     description     = description)     
        self.setState(state='')

        

Trait.known_traits.append(CloneTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class MovedTrait(Trait):
    ID = 'markmoved'
    def __init__(self,
                 image      = 'moved.gif',
                 xoff       = 36,
                 yoff       = -38,
                 name       = 'Mark moved',
                 key        = key('M'),
                 dummy      = ''  # Description
                 # ignoreSame = True
                 ):
        '''Create a moved trait (VASSAL.counters.MovementMarkable)'''
        super(MovedTrait,self).__init__()
        self.setType(image    = image,
                     xoff     = xoff,
                     yoff     = yoff,
                     name     = name,
                     key      = key,
                     dummy    = dummy, # Description
                     # ignoreSame = ignoreSame
                     )
        self.setState(moved = False)

Trait.known_traits.append(MovedTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class TrailTrait(Trait):
    ID = 'footprint'
    def __init__(self,
                 key             = key('T'),
                 name            = 'Movement Trail',
                 localVisible    = False, # Start on
                 globalVisible   = True, # Visible to all players
                 radius          = 10,
                 fillColor       = rgb(255,255,255),
                 lineColor       = rgb(0,0,0),
                 activeOpacity   = 100,
                 inactiveOpacity = 50,
                 edgesBuffer     = 20,
                 displayBuffer   = 30,
                 lineWidth       = 5,
                 turnOn          = key(NONE,0)+',wgTrailsOn',
                 turnOff         = key(NONE,0)+',wgTrailsOff',
                 reset           = '',
                 description     = 'Enable or disable movement trail'):        
        ''' Create a movement trail trait ( VASSAL.counters.Footprint)'''
        super(TrailTrait,self).__init__()
        lw = (lineWidth
              if isinstance(lineWidth,str) and lineWidth.startswith('{') else
              int(lineWidth))
        ra = (radius
              if isinstance(radius,str) and radius.startswith('{') else
              int(radius))
        
        self.setType(key               = key,# ENABLE KEY
                     name              = name,# MENU 
                     localVisible      = localVisible,# LOCAL VISABLE
                     globalVisible     = globalVisible,# GLOBAL VISABLE
                     radius            = ra,# RADIUS
                     fillColor         = fillColor,# FILL COLOR
                     lineColor         = lineColor,# LINE COLOR 
                     activeOpacity     = activeOpacity,# ACTIVE OPACITY
                     inactiveOpacity   = inactiveOpacity,# INACTIVE OPACITY
                     edgesBuffer       = edgesBuffer,# EDGES BUFFER
                     displayBuffer     = displayBuffer,# DISPLAY BUFFER
                     lineWidth         = lw,# LINE WIDTH 
                     turnOn            = turnOn,# TURN ON KEY
                     turnOff           = turnOff,# TURN OFF KEY
                     reset             = reset,# RESET KEY
                     description       = description)       # DESC
        self.setState(isGlobal  = False,
                      map       = '',
                      points    = 0,     # POINTS (followed by [; [X,Y]*]
                      init      = False) 

Trait.known_traits.append(TrailTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

class LabelTraitCodes:
    TOP    = 't'
    BOTTOM = 'b'
    CENTER = 'c'
    LEFT   = 'l'
    RIGHT  = 'r'
    PLAIN  = 0
    BOLD   = 1
    ITALIC = 2
    
# --------------------------------------------------------------------
class LabelTrait(Trait):
    ID     = 'label'
    def __init__(self,
                 label           = None,
                 labelKey        = '',
                 menuCommand     ='Change label',
                 fontSize        = 10,
                 background      = 'none',
                 foreground      = '255,255,255',
                 vertical        = LabelTraitCodes.TOP,
                 verticalOff     = 0,
                 horizontal      = LabelTraitCodes.CENTER,
                 horizontalOff   = 0,
                 verticalJust    = LabelTraitCodes.BOTTOM,
                 horizontalJust  = LabelTraitCodes.CENTER,
                 nameFormat      = '$pieceName$ ($label$)',
                 fontFamily      = 'Dialog',
                 fontStyle       = LabelTraitCodes.PLAIN,
                 rotate          = 0,
                 propertyName    = 'TextLabel',
                 description     = '',
                 alwaysUseFormat = False):
        '''Create a label trait (can be edited property)

        Note that rotation comes last in the operations.  That is,
        `horizontal...` and `vertical...` must be specified as if the
        label is not rotated, and then rotation is applied.

        Negative vertical offset moves the label _up_. 

        '''
        super(LabelTrait,self).__init__()
        if not background or background == 'none': background = ''
        if not foreground or foreground == 'none': foreground = ''
        self.setType(labelKey		= labelKey,
                     menuCommand	= menuCommand,
                     fontSize		= fontSize,
                     background		= background,
                     foreground		= foreground,
                     vertical		= vertical,
                     verticalOff	= verticalOff,
                     horizontal		= horizontal,
                     horizontalOff	= horizontalOff,
                     verticalJust	= verticalJust,
                     horizontalJust	= horizontalJust,
                     nameFormat		= nameFormat,
                     fontFamily		= fontFamily,
                     fontStyle		= fontStyle,
                     rotate		= rotate,
                     propertyName	= propertyName,
                     description	= description,
                     alwaysUseFormat	= alwaysUseFormat)
        self.setState(label = (nameFormat if label is None else label))


Trait.known_traits.append(LabelTrait)

#
# EOF
#



#!/usr/bin/env python
#
# CyberBoard GitHub page:
#
#    https://github.com/CyberBoardPBEM/cbwindows
#
#
# ====================================================================
## BEGIN_IMPORT
from pywargame.cyberboard.scenario  import Scenario
from pywargame.cyberboard.extractor import GSNExtractor
from pywargame.common               import Verbose
## END_IMPORT
    

# ====================================================================
def extractMain():
    from argparse import ArgumentParser, FileType
    from pathlib import Path

    ap = ArgumentParser(description='Extract information from a '
                        'CyberBoard Scenario file and store in ZIP archive',
                        epilog='Default output file name is input file name '
                        'with .gsn replaced by .zip')
    ap.add_argument('input', type=str, help='The file')
    ap.add_argument('output',type=str, nargs='?',help='Output',default='')
    ap.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')
    ap.add_argument('-D','--dump',type=str,nargs='*',default='',
                    help='Dump content')
    ap.add_argument('-X','--gamebox',
                    type=str,
                    default=None,
                    help='Override gamebox (.gbx) file')

    args = ap.parse_args()
    Verbose().setVerbose(args.verbose)
    
    gsn = Scenario.fromFile(args.input,args.gamebox)
    rat = GSNExtractor(gsn)

    out = args.output
    if out == '':
        p = Path(args.input)
        out = p.stem+'.zip'
    rat.save(out)

    if 'save' in args.dump:
        print(rat)
    if 'gsn' in args.dump:
        print(gsn)

# ====================================================================
if __name__ == '__main__':
    extractMain()
    
#
# EOF
#

    

    



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class ReturnTrait(Trait):
    ID      = 'return'
    def __init__(self,
                 command         = '',
                 key             = '',
                 deckId          = '',
                 prompt          = '',
                 description     = '',
                 version         = 2,
                 select          = False, # If true, select at run-time
                 expression      = ''
                 ):
        '''Create a return trait (VASSAL.counter.ReturnToDeck)'''
        super().__init__()

        self.setType(command         = command,          # Context menu name
                     key             = key,              # Context menu key
                     deckId          = deckId,
                     prompt          = prompt,
                     description     = description,
                     version         = version,
                     select          = select,
                     expression      = expression
                     )     
        self.setState(state='')

        

Trait.known_traits.append(ReturnTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class LayerTrait(Trait):
    ID = 'emb2'
    def __init__(self,
                 images       = [''],
                 newNames     = None,
                 activateName = 'Activate',
                 activateMask = CTRL,
                 activateChar = 'A',
                 increaseName = 'Increase',
                 increaseMask = CTRL,
                 increaseChar = '[',
                 decreaseName = '',
                 decreaseMask = CTRL,
                 decreaseChar  = ']',
                 resetName    = '',
                 resetKey     = '',
                 resetLevel   = 1,
                 under        = False,
                 underXoff    = 0,
                 underYoff    = 0,
                 loop         = True,
                 name         = '',
                 description  = '',
                 randomKey    = '',
                 randomName   = '',
                 follow       = False,
                 expression   = '',
                 first        = 1,
                 version      = 1, # 1:new, 0:old
                 always       = True,
                 activateKey  = key('A'),
                 increaseKey  = key('['),
                 decreaseKey  = key(']'),
                 scale        = 1.):
        '''Create a layer trait (VASSAL.counter.Embellishment)'''
        super(LayerTrait,self).__init__()
        if newNames is None and images is not None:
            newNames = ['']*len(images)
        self.setType(
            activateName        = activateName,
            activateMask        = activateMask,
            activateChar        = activateChar,
            increaseName        = increaseName,
            increaseMask        = increaseMask,
            increaseChar        = increaseChar,
            decreaseName        = decreaseName,
            decreaseMask        = decreaseMask,
            decreaseChar        = decreaseChar,
            resetName           = resetName,
            resetKey            = resetKey,
            resetLevel          = resetLevel,
            under               = under,
            underXoff           = underXoff,
            underYoff           = underYoff,
            images              = ','.join(images),
            newNames            = ','.join(newNames),
            loop                = loop,
            name                = name,
            randomKey           = randomKey,
            randomName          = randomName,
            follow              = follow,
            expression          = expression,
            first               = first,
            version             = version,
            always              = always,
            activateKey         = activateKey,
            increaseKey         = increaseKey,
            decreaseKey         = decreaseKey,
            description         = description,
            scale               = scale)
        self.setState(level=1)

    def getImages(self):
        return self['images'].split(',')

    def setImages(self,*images):
        self['images'] = ','.join(images)
        
    def getNames(self):
        return self['newNames'].split(',')

    def setNames(self,*names):
        images   = self.getImages()
        newNames = ','.join(names+['']*(len(images)-len(names)))
        self['newNames'] = newNames

    def getLevel(self):
        return self['level']

    def setLevel(self,level):
        self['level'] = level
                            

Trait.known_traits.append(LayerTrait)

#
# EOF
#



## BEGIN_IMPORT
from . vmod import VMod
from .. common import VerboseGuard, Verbose 
## END_IMPORT

class Exporter:
    def __init__(self):
        '''Base class for exporters'''
        pass


    def setup(self):
        '''Should be defined to set-up for processing, for example
        generating images and such.  This is executed in a context
        where the VMod file has been opened for writing via
        `self._vmod`. Thus, files can be added to the module at this
        stage.
        '''         
        pass

    def createBuildFile(self,ignores='(.*markers?|all|commons|[ ]+)'):
        '''Should be defined to make the `buildFile.xml` document

        Parameters
        ----------
        ignores : str
            Regular expression to match ignored categories for factions
            determination. Python's re.fullmatch is applied to this
            regular exression against chit categories.  If the pattern
            is matched, then the chit is not considered to belong to a
            faction.

        '''
        pass

    def createModuleData(self):
        '''Should be defined to make the `moduledata` document'''
        pass
    
    def run(self,vmodname,patch=None):
        '''Run the exporter to generate the module
        '''
        with VMod(vmodname,'w') as vmod:
            self._vmod = vmod
            self.setup()
            self.createBuildFile() 
            self.createModuleData()
            self.runPatch(patch)
            self._vmod.addFiles(**{VMod.BUILD_FILE  :
                                   self._build.encode(),
                                   VMod.MODULE_DATA :
                                   self._moduleData.encode()})
        Verbose().message('Created VMOD')
        

    def runPatch(self,patch):
        '''Run user specified patch script.  The script should define

            ```
            def patch(buildFile,moduleData,vmod,verbose):
            ```

        where `buildFile` is the `buildFile.xml` and `moduleData` are
        the XML documents as `xml.dom.Document` objects, `vmod` is a
        `VMod` instance, and `verbose` is a `bool` selecting verbose
        mode or not.
        '''
        if patch is None or patch == '':
            return
        
        from importlib.util import spec_from_file_location, module_from_spec
        from pathlib import Path
        from sys import modules

        p = Path(patch)
        with VerboseGuard(f'Will patch module with {p.stem}.patch function') \
             as v:

            spec   = spec_from_file_location(p.stem, p.absolute())
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
            modules[p.stem] = module
            
            # Patch must accept xml.dom.document,xml.dom.document,ZipFile
            module.patch(self._build,
                         self._moduleData,
                         self._vmod,
                         Verbose().verbose)
    



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class SheetTrait(Trait):
    ID             = 'propertysheet'
    EVERY_KEY_TEXT = 'Every Keystroke'
    APPLY_TEXT     = 'Apply Button or Enter Key'
    CLOSE_TEXT     = 'Close Window or Enter Key'
    EVERY_KEY      = 0
    APPLY          = 1
    CLOSE          = 2
    TEXT           = 0
    AREA           = 1
    LABEL          = 2
    TICKS          = 3
    TICKS_MAX      = 4
    TICKS_VAL      = 5
    TICKS_BOTH     = 6
    SPINNER        = 7
    TYPE_DELIM     = ';'
    DEF_DELIM      = '~'
    STATE_DELIM    = '~'
    LINE_DELIM     = '|'
    VALUE_DELIM    = '/'

    @classmethod 
    def encodeState(cls,k,e):
        type = e['type']
        if type == cls.TEXT:    return f'{e["value"]}'
        if type == cls.AREA:    return f'{e["value"].replace("\n",LINE_DELIM)}'
        if type == cls.SPINNER: return f'{e["value"]}'
        if type in [cls.TICKS,cls.TICKS_MAX,cls.TICKS_VAL,cls.TICKS_BOTH]:
            try:
                val = int(e["value"])
            except:
                val = 0
            try:
                max = int(e["max"])
            except:
                max = 0
            return f'{val}{cls.VALUE_DELIM}{max}'
        return ''

    @classmethod
    def encodeDefinition(cls,rows):
        definition = cls.DEF_DELIM.join([f'{e["type"]}{k}'
                                     for k,e in rows.items()])
                                
        state      = cls.STATE_DELIM.join([self.encodeState(k,e)
                                       for k,e in rows.items()])
        return definition, state

    @classmethod
    def decodeDefinition(cls,definitions,state):
        tns  = definitions.split(cls.DEF_DELIM)
        sts  = state      .split(cls.STATE_DELIM)
        def decodeDef(d):
            try:
                type = int(d[0])
            except:
                type = cls.TEXT
            return type, d[1:]
        rows = {}
        for tn, st in zip(tns,sts):
            type, name = decodeDef(tn)
            rows[name] = { 'type': type }
            rows[name].update(cls.decodeState(name,type,st))
        return rows

    @classmethod 
    def decodeState(cls,name,type,state):
        if type == cls.TEXT:    return {'value': state}
        if type == cls.AREA:    return {'value': state.replace('|','\n') }
        if type == cls.SPINNER: return {'value': state}
        if type in [cls.TICKS,
                    cls.TICKS_MAX,
                    cls.TICKS_VAL,
                    cls.TICKS_BOTH]:
            fields = state.split(VALUE_DELIM)
            try:
                value = int(fields[0])
            except:
                value = 0
            try:
                max = int(fields[1])
            except:
                max = 0
            return {'value': value, 'max': max }
        return {}
    
    def __init__(self,
                 command         = '',
                 commit          = EVERY_KEY,
                 color           = ',,',
                 key             = '',
                 description     = '',
                 rows            = {}):
        '''Create a clone trait (VASSAL.counter.Clone)'''
        super().__init__()
        definition, state = self.encodeDefinition(rows)
        rgbcol = color.split(',')
        self.setType(definition      = definition,
                     command         = command,          # Context menu name
                     letter          = '',
                     commit          = commit,
                     red             = rgbcol[0],
                     green           = rgbcol[1],
                     blue            = rgbcol[2],
                     key             = key,              # Context menu key
                     description     = description)     
        self.setState(state=state)
        
        
    def getDefinitionState(self):
        return self.decodeDefinition(self['definition'],self['state'])

    def setDefinitionState(self,rows):
        self['definition'], self['state'] = self.encodeDefinition(rows)
    

Trait.known_traits.append(SheetTrait)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . game import Game
from . xmlns import xmlns
## END_IMPORT

# --------------------------------------------------------------------
class BuildFile(Element):
    def __init__(self,root=None):
        '''Construct from a DOM object, if given, otherwise make new'''
        # from xml.dom.minidom import Document
        super(BuildFile,self).__init__(None,'',None)
        
        self._root = root
        self._tag  = 'buildFile'
        if self._root is None:
            self._root = xmlns.Document()

        self._node = self._root

    def addGame(self,**kwargs):
        '''Add a `Game` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Game
            The added element
        '''
        return Game(self,**kwargs)

    def getGame(self):
        '''Get the `Game`'''
        try:
            return Game(self,
                        node=self._root.\
                        getElementsByTagName('VASSAL.build.GameModule')[0])
        except:
            pass

        return Game(self,
                    node=self._root.\
                    getElementsByTagName('VASSAL.launch.BasicModule')[0])
                    

    def encode(self):
        '''Encode into XML'''
        return self._root.toprettyxml(indent=' ',
                                      encoding="UTF-8",
                                      standalone=False)


#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import CbManager
## END_IMPORT

# ====================================================================
class GBXMarkDef:
    def __init__(self,ar):
        '''Definition of a mark'''
        with VerboseGuard(f'Reading mark definition'):
            self._id    = ar.iden()
            self._flags = ar.word()

    def __str__(self):
        return f'Mark: {self._id:04x},{self._flags:04x}'
    
# --------------------------------------------------------------------
class GBXMarkSet:
    def __init__(self,ar):
        '''Set of marks'''
        with VerboseGuard(f'Reading mark set'):
            self._name   = ar.str()
            self._viz    = ar.word()
            n            = ar.sub_size()
            self._marks = [ar.iden() for _ in range(n)]

    def __len__(self):
        return len(self._marks)

    def __str__(self):
        return (f'{self._name}: '+','.join([str(p) for p in self._marks]))
    
# --------------------------------------------------------------------
class GBXMarkManager:
    def __init__(self,ar):
        '''Manager of marks'''
        with VerboseGuard(f'Reading mark manager'):
            self._reserved     = [ar.word() for _ in range(4)] 
            n                  = ar.iden()
            self._marks        = [GBXMarkDef(ar) for _ in range(n)]
            n                  = ar.sub_size()
            self._sets         = [GBXMarkSet(ar) for _ in range(n)]
        
    def __len__(self):
        return len(self._sets)

    def toDict(self,tileManager,strings):
        from math import log10, ceil
        with VerboseGuard(f'Creating dict from mark manager'):
            setDigits  = int(ceil(log10(len(self)+.5)))
            markDigits = 1
            for markSet in self._sets:
                markDigits = max(markDigits,
                                 int(ceil(log10(len(markSet)+.5))))

            marksMap  = {}
            setList   = []
            ret = {'map': marksMap,
                   'sets': setList }
            
            for ips, markSet in enumerate(self._sets):
                with VerboseGuard(f'Creating dict mark set {markSet._name}'):
                    setPrefix = f'mark_{ips:0{setDigits}d}'
                    idList    = []
                    setDict   = { 'description': markSet._name.strip(),
                                  'marks':       idList }
                    setList.append(setDict)
                    
                    for ipc, markID in enumerate(markSet._marks):
                        markPrefix = f'{setPrefix}_{ipc:0{markDigits}d}'
                        markDef    = self._marks[markID]
                        tmpStr     = strings._id2str.get(markID|0xF0000000,'')
                        markDesc   = tmpStr.replace('\r','').replace('\n',', ')
                        markDict   = {}
                        if markDesc != '':
                            markDict['description'] = markDesc
                        marksMap[markID] = markDict
                        idList  .append(markID)
                        
                        img = tileManager.image(markDef._id)
                    
                        if img is None:
                            continue
                    
                        sav     = f'{markPrefix}.png'
                        markDict.update({'image':     img,
                                         'filename':  sav })

            return ret
    
    def __str__(self):
        return ('Mark manager:\n'
                +f'Reserved: {self._reserved}\n'
                +f'# marks: {len(self._marks)}\n  '
                +'\n  '.join([str(p) for p in self._marks])+'\n'
                +f'# mark sets: {len(self._sets)}\n  '
                +'\n  '.join([str(p) for p in self._sets])
                )

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . head import num_version, readVector
from . features import Features
## END_IMPORT

class GSNPalette:
    def __init__(self,ar):
        with VerboseGuard('Reading palette'):
            self._visible    = ar.word()
            self._comboIndex = ar.dword()
            self._topIndex   = ar.dword()
            
class GSNTrayPalette(GSNPalette):
    def __init__(self,ar,vers,iden):
        with VerboseGuard(f'Reading scenario tray palette {iden}'):
            super(GSNTrayPalette,self).__init__(ar)
            self._iden       = iden
            self._listSel    = readVector(ar,lambda ar : ar.dword())
            
    def __str__(self):
        return f'GSNTrayPalette: {self._comboIndex} '\
            f'{self._topIndex} {self._listSel}\n'

class GSNMarkPalette(GSNPalette):
    def __init__(self,ar,vers):
        with VerboseGuard(f'Reading scenario mark palette'):
            super(GSNMarkPalette,self).__init__(ar)
            self._listSel    = ar.dword()
    def __str__(self):
        return f'GSNMarkPalette: {self._comboIndex} '\
            f'{self._topIndex} {self._listSel}\n'

#
# EOF
#




## BEGIN_IMPORT
from . base import *
from . element import ToolbarElement
## END_IMPORT

# --------------------------------------------------------------------
class GlobalKey(ToolbarElement):
    ## Fast filters
    ##
    ## Format:
    ##
    ##   GKCTYPE|LOCATION|<location>|PROPERTY|<property>
    ##
    ## where
    ##
    ##    GKCTYPE: MAP    (from Global),
    ##             MODULE (from GameGlobal)
    ##             DECK
    ##             COUNTER (from CounterGlobal)
    ##
    ##    LOCATION: true - filter on location
    ##              false - do not filter on location
    ##
    ##    PROPERTY: true - filter on property
    ##              false - do not filter on property
    ##
    ##    <location> is
    ##
    ##       TARGETTYPE|TARGETMAP|TARGETBOARD|TARGETZONE|TARGETLOCATION|TARGETX|TARGETY|TARGETDECK
    ##
    ##    <property> is
    ##
    ##       PROPERTYNAME|PROPERTYVALUE|COMPARE
    ##
    ##    TARGETTYPE: CURSTACK current stack or deck
    ##                CURMAP   current map
    ##                CURZONE  current zone
    ##                CURLOC   current location 
    ##                MAP      specified map      (TARGETMAP)
    ##                ZONE     specified zone     (TARGETZONE)
    ##                LOCATION specified location (TARGETLOCATION)
    ##                XY       specified X,Y      (TARGETBOARD,TARGETX,TARGETY)
    ##                DECK     specified deck     (TARGETDECK)
    ##                CURMAT   current mat
    ##                
    ##    PROPERTYNAME
    ##
    ##    PROPERTYVALUE: A constant
    ##    
    ##    COMPARE: EQUALS:         ==
    ##             NOT_EQUALS:     !=
    ##             GREATER:        >
    ##             GREATER_EQUALS: >=
    ##             LESS:           <
    ##             LESS_EQUALS:    <=
    ##             MATCH:          =~
    ##             NOT_MATCH       !~
    ##
    ## 
    SELECTED = 'MAP|false|MAP|||||0|0||true|Selected|true|EQUALS'
    UNIQUE = ['name']
    def __init__(self,
                 parent,
                 tag,
                 node                 = None,
                 name                 = '',                
                 icon                 = '',
                 tooltip              = '',
                 buttonHotkey         = '',
                 buttonText           = '',
                 canDisable           = False,
                 propertyGate         = '',
                 disabledIcon         = '',
                 # Local
                 hotkey               = '',
                 deckCount            = '-1',
                 filter               = '',
                 reportFormat         = '',
                 reportSingle         = False,
                 singleMap            = True,
                 target               = SELECTED):
        '''
        Parameters
        ----------
        - tag          The XML tag to use
        - parent       Parent node
        - node         Optionally existing node
        - name         Name of key
        - buttonHotkey Key in "global" scope
        - hotkey       Key to send to targeted pieces
        - buttonText   Text on button
        - canDisable   If true, disabled when propertyGate is true
        - deckCount    Number of decks (-1 is all)
        - filter       Which units to target
        - propertyGate When true, disable
        - reportFormat Chat message
        - reportSingle Also show single piece reports
        - singleMap    Only originating map if True
        - target       Preselection filter (default selected pieces)
        - tooltip      Hover-over message
        - icon         Image to use as icon

        Default targets are selected units
        '''
        
        super(GlobalKey,self).\
            __init__(parent,
                     tag,
                     node                 = node,
                     name                 = name,
                     icon                 = icon,
                     tooltip              = tooltip,
                     buttonHotkey         = buttonHotkey, # This hot key
                     buttonText           = buttonText,
                     canDisable           = canDisable,
                     propertyGate         = propertyGate,
                     disabledIcon         = disabledIcon,
                     hotkey               = hotkey,       # Target hot key
                     deckCount            = deckCount,
                     filter               = filter,
                     reportFormat         = reportFormat,
                     reportSingle         = reportSingle,
                     singleMap            = singleMap,
                     target               = target)
#
# EOF
# 



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . grid import *
## END_IMPORT

# --------------------------------------------------------------------
class ZonedGrid(Element):
    TAG=Element.BOARD+'ZonedGrid'
    def __init__(self,board,node=None):
        super(ZonedGrid,self).__init__(board,self.TAG,node=node)

    def getBoard(self):
        ## BEGIN_IMPORT
        from . board import Board
        ## END_IMPORT
        b = self.getParent(Board)
        # print(f'Get Board of Zoned: {b}')        
        return b
    def getPicker(self):
        z = self.getBoard()
        if z is not None:
            return z.getPicker()
        return None
    def getMap(self):
        z = self.getPicker()
        if z is not None:
            return z.getMap()
        return None
    def addHighlighter(self,**kwargs):
        '''Add a `Highlighter` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Highlighter
            The added element
        '''
        return self.add(ZonedGridHighlighter,**kwargs)
    def getHighlighters(self,single=True):
        '''Get all or a sole `ZonedGridHighlighter` element(s) from this
        
        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Highligter` child, otherwise fail.
            If `False` return all `Highligter` children in this element
        
        Returns
        -------
        children : list
            List of `Highligter` children (even if `single=True`)
        '''
        return self.getAllElements(ZonedGridHighlighter,single=single)
    def addZone(self,**kwargs):
        '''Add a `Zone` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Zone
            The added element
        '''
        return self.add(Zone,**kwargs)
    def getZones(self,asdict=True):
        '''Get all Zone element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Zone` elements.  If `False`, return a list of all Zone` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Zone` children
        '''
        return self.getElementsByKey(Zone,'name',asdict=asdict)

registerElement(ZonedGrid)

# --------------------------------------------------------------------
class ZonedGridHighlighter(Element):
    TAG=Element.BOARD+'mapgrid.ZonedGridHighlighter'
    def __init__(self,zoned,node=None):
        super(ZonedGridHighlighter,self).__init__(zoned,self.TAG,node=node)
    def getZonedGrid(self): return self.getParent(ZonedGrid)

    def addZoneHighlight(self,**kwargs):
        '''Add a `ZoneHighlight` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : ZoneHighlight
            The added element
        '''
        return self.add(ZoneHighlight,**kwargs)
    def getZoneHighlights(self,asdict=True):
        '''Get all ZoneHighlight element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Zone` elements.  If `False`, return a list of all Zone` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Zone` children
        '''
        return self.getElementsByKey(ZoneHighlight,'name',asdict=asdict)

registerElement(ZonedGridHighlighter)

# --------------------------------------------------------------------
class ZoneHighlight(Element):
    TAG=Element.BOARD+'mapgrid.ZoneHighlight'
    FULL    = 'Entire Zone',
    BORDER  = 'Zone Border',
    PLAIN   = 'Plain',
    STRIPED = 'Striped'
    CROSS   = 'Crosshatched',
    TILES   = 'Tiled Image'
    UNIQUE  = ['name']
    def __init__(self,
                 highlighters,
                 node     = None,
                 name     = '',
                 color    = rgb(255,0,0),
                 coverage = FULL,
                 width    = 1,
                 style    = PLAIN,
                 image    = '',
                 opacity  = 50):
        super(ZoneHighlight,self).__init__(highlighters,
                                           self.TAG,
                                           node     = node,
                                           name     = name,
                                           color    = color,
                                           coverage = coverage,
                                           width    = width,
                                           style    = style,
                                           image    = image,
                                           opacity  = int(opacity))
    def getZonedGridHighlighter(self):
        return self.getParent(ZonedGridHighlighter)


registerElement(ZoneHighlight)


# --------------------------------------------------------------------
class ZoneProperty(Element):
    TAG = Element.MODULE+'properties.ZoneProperty'
    UNIQUE  = ['name']
    def __init__(self,zone,node=None,
                 name         = '',
                 initialValue = '',
                 isNumeric    = False,
                 min          = "null",
                 max          = "null",
                 wrap         = False,
                 description  = ""):
        super(ZoneProperty,self).__init__(zone,self.TAG,
                                            node         = node,
                                            name         = name,
                                            initialValue = initialValue,
                                            isNumeric    = isNumeric,
                                            min          = min,
                                            max          = max,
                                            wrap         = wrap,
                                            description  = description)

    def getZone(self):
        return self.getParent(Zone)

registerElement(ZoneProperty)

# --------------------------------------------------------------------
class Zone(Element):
    TAG = Element.BOARD+'mapgrid.Zone'
    UNIQUE  = ['name']
    def __init__(self,zoned,node=None,
                 name              = "",
                 highlightProperty = "",
                 locationFormat    = "$gridLocation$",
                 path              = "0,0;976,0;976,976;0,976",
                 useHighlight      = False,
                 useParentGrid     = False):
        super(Zone,self).\
            __init__(zoned,self.TAG,node=node,
                     name              = name,
                     highlightProperty = highlightProperty,
                     locationFormat    = locationFormat,
                     path              = path,
                     useHighlight      = useHighlight,
                     useParentGrid     = useParentGrid)

    def getZonedGrid(self):
        z = self.getParent(ZonedGrid)
        # print(f'Get Zoned of Zone {self["name"]}: {z}')        
        return z
    
    def getBoard(self):
        z = self.getZonedGrid()
        if z is not None:
            return z.getBoard()
        return None
    def getPicker(self):
        z = self.getBoard()
        if z is not None:
            return z.getPicker()
        return None
    def getMap(self):
        z = self.getPicker()
        if z is not None:
            return z.getMap()
        return None    
    def addHexGrid(self,**kwargs):
        '''Add a `HexGrid` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : HexGrid
            The added element
        '''
        return self.add(HexGrid,**kwargs)
    def addSquareGrid(self,**kwargs):
        '''Add a `SquareGrid` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : SquareGrid
            The added element
        '''
        return self.add(SquareGrid,**kwargs)
    def addRegionGrid(self,**kwargs):
        '''Add a `RegionGrid` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : RegionGrid
            The added element
        '''
        return self.add(RegionGrid,**kwargs)
    def addProperty(self,**kwargs):
        '''Add a `Property` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Property
            The added element
        '''
        return self.add(ZoneProperty,**kwargs)
    def getHexGrids(self,single=True):
        '''Get all or a sole `HexGrid` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `HexGrid` child, otherwise fail.
            If `False` return all `HexGrid` children in this element
        
        Returns
        -------
        children : list
            List of `HexGrid` children (even if `single=True`)
        '''
        return self.getAllElements(HexGrid,single=single)
    def getSquareGrids(self,single=True):
        '''Get all or a sole `SquareGrid` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `SquareGrid` child, otherwise fail.
            If `False` return all `SquareGrid` children in this element
        
        Returns
        -------
        children : list
            List of `SquareGrid` children (even if `single=True`)
        '''
        return self.getAllElements(SquareGrid,single=single)
    def getRegionGrids(self,single=True):
        '''Get all or a sole `RegionGrid` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `RegionGrid` child, otherwise fail.
            If `False` return all `RegionGrid` children in this element
        
        Returns
        -------
        children : list
            List of `RegionGrid` children (even if `single=True`)
        '''
        return self.getAllElements(RegionGrid,single=single)
    def getGrids(self,single=True):
        '''Get all or a sole `Grid` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Grid` child, otherwise fail.
            If `False` return all `Grid` children in this element
        
        Returns
        -------
        children : list
            List of `Grid` children (even if `single=True`)
        '''
        g = self.getHexGrids(single=single)
        if g is not None: return g

        g = self.getSquareGrids(single=single)
        if g is not None: return g

        g = self.getRegionGrids(single=single)
        if g is not None: return g

        return g
    def getProperties(self):
        '''Get all `Property` element from this

        Returns
        -------
        children : dict
            dict of `Property` children
        '''
        return getElementsByKey(ZoneProperty,'name')
    
    def getPath(self):
        p  = self['path'].split(';')
        r  = []
        for pp in p:
            c = pp.split(',')
            r.append([int(c[0]),int(c[1])])
        return r
    
    def getBB(self):
        from functools import reduce
        path = self.getPath()
        llx  = reduce(lambda old,point:min(point[0],old),path,100000000000)
        lly  = reduce(lambda old,point:min(point[1],old),path,100000000000)
        urx  = reduce(lambda old,point:max(point[0],old),path,-1)
        ury  = reduce(lambda old,point:max(point[1],old),path,-1)
        return llx,lly,urx,ury
    def getWidth(self):
        llx,_,urx,_ = self.getBB()
        return urx-llx
    def getHeight(self):
        _,lly,_,ury = self.getBB()
        return ury-lly
    def getXOffset(self):
        return self.getBB()[0]
    def getYOffset(self):
        return self.getBB()[1]

registerElement(Zone)

#
# EOF
#



#!/usr/bin/env python
import sys
sys.path.append('..')

from common import Verbose, VerboseGuard
from common.collector import Collector

        
if __name__ == '__main__':
    from argparse import ArgumentParser, FileType  

    ap = ArgumentParser(description='Collect to single script')
    ap.add_argument('output',
                    type=FileType('w'),
                    nargs='?',
                    default='vassal.py',
                    help='Output script name')
    ap.add_argument('-v','--verbose',action='store_true',
                    help='Be verbose')

    args = ap.parse_args()

    Verbose().setVerbose(args.verbose)

    c = Collector(executable=False)
    c.run(args.output,
          '../common/singleton.py',
          '../common/verbose.py',
          '../common/verboseguard.py',
          '../common/dicedraw.py',
          'xmlns.py',
          'base.py',
          'element.py',
          'folder.py',
          'globalkey.py',
          'gameelements.py',
          'mapelements.py',
          'globalproperty.py',
          'turn.py',
          'documentation.py',
          'player.py',
          'chessclock.py',
          'widget.py',
          'grid.py',
          'zone.py',
          'board.py',
          'map.py',
          'chart.py',
          'command.py',
          'trait.py',
          'withtraits.py',
          'extension.py',
          'traits/area.py',
          'traits/clone.py',
          'traits/dynamicproperty.py',
          'traits/globalproperty.py',
          'traits/prototype.py',
          'traits/place.py',
          'traits/report.py',
          'traits/calculatedproperty.py',
          'traits/restrictcommand.py',
          'traits/label.py',
          'traits/layer.py',
          'traits/globalcommand.py',
          'traits/globalhotkey.py',
          'traits/nostack.py',
          'traits/deselect.py',
          'traits/restrictaccess.py',
          'traits/rotate.py',
          'traits/stack.py',
          'traits/mark.py',
          'traits/mask.py',
          'traits/trail.py',
          'traits/delete.py',
          'traits/sendto.py',
          'traits/moved.py',
          'traits/skel.py',
          'traits/submenu.py',
          'traits/basic.py',
          'traits/trigger.py',
          'traits/nonrect.py',
          'traits/click.py',
          'traits/mat.py',
          'traits/cargo.py',
          'traits/movefixed.py',
          'traits/sheet.py',
          'traits/hide.py',
          'traits/retrn.py',
          'game.py',
          'buildfile.py',
          'moduledata.py',
          'save.py',
          'vsav.py',
          'vmod.py',
          'upgrade.py',
          'exporter.py',
          'merger.py')

    



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . gameelements import GameElement
## END_IMPORT

# ====================================================================
class ChessClock(Element):
    TAG=Element.MODULE+'chessclockcontrol.ChessClock'
    UNIQUE = ['side']
    def __init__(self,
                 doc,
                 node                   = None,
                 icon                   = '',
                 description            = '',
                 side                   = '',
                 tooltip                = 'Individual clock control',
                 buttonText             = '',
                 startHotkey            = '',
                 stopHotkey             = '',
                 tickingBackgroundColor = rgb(255,255,0),
                 tickingFontColor       = rgb(0,0,0),
                 tockingFontColor       = rgb(51,51,51)):
        '''Individual clock for a side

        When the clock is running, the background colour may be
        changed, and the colour of the numbers alternate between
        `tickingFontColor` and `tockingFontColor`.
        
        Parameters
        ----------
        doc : Element
            Parent element 
        node : xml.dom.Element 
            Read from this node
        icon : str
            File name of button icon
        description : str
            Note on this clock
        side : str
            Name of side this clock belongs to
        tooltop : str
            Hover help text
        buttonText : str
            Text on button
        startHotkey : str (key code)
            Key or command to start timer
        stopHotkey : str (key code)
            Key or command to stop timer
        tickingBackgroundColor : str (color)
            Background color of time display when clock is running
        tickingFontColor : str (color)
            First color of numbers in display when clock is running.
        tockingFontColor : str (color)
            Second color of numbers in display when clock is running.
        '''
        super(ChessClock,self).__init__(#ChessClock
                 doc,
                 self.TAG,
                 node                   = node,
                 icon                   = icon,
                 description            = description,
                 side                   = side,
                 tooltip                = tooltip,
                 buttonText             = buttonText,
                 startHotkey            = startHotkey,
                 stopHotkey             = stopHotkey,
                 tickingBackgroundColor = tickingBackgroundColor,
                 tickingFontColor       = tickingFontColor,
                 tockingFontColor       = tockingFontColor)
            
    def getControl(self):
        '''Get Parent element'''
        return self.getParent(ChessClockControl)

registerElement(ChessClock)

# ====================================================================
class ChessClockControl(GameElement):
    TAG=Element.MODULE+'ChessClockControl'
    ALWAYS = 'Always'
    AUTO   = 'Auto'
    NEVER  = 'Never'
    UNIQUE = ['name']
    def __init__(self,
                 doc,
                 node              = None,
                 name              = 'Chess clock',
                 description       = '',
                 buttonIcon        = 'chess_clock.png',
                 buttonText        = '',
                 buttonTooltip     = 'Show/stop/hide chess clocks',
                 showHotkey        = key('U',ALT),
                 pauseHotkey       = key('U',CTRL_SHIFT),
                 nextHotkey        = key('U'),
                 startOpponentKey  = '',
                 showTenths        = AUTO,
                 showSeconds       = AUTO,
                 showHours         = AUTO,
                 showDays          = AUTO,
                 allowReset        = False,
                 addClocks         = True):
        '''A set of chess clocs

        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        name : str
            Name of clock control
        description : str
            Note on the chess clocks control
        buttonIcon : str
            Icon file name for button (chess_clock.png)
        buttonText : str
            Text on button 
        buttonTooltip : str
            Hower help
        showHotkey : str (key code)
            Show or hide interface hot key
        nextHotkey : str (key code)
            Start the next clock hot key
        pauseHotkey : str (key code)
            Pause all clocks hot key 
        startOpponentKey : str (key code)
            Start opponens clock 
        showTenths : one of AUTO, ALWAYS, NEVER
            Whether to show tenths of seconds
        showSeconds : one of AUTO, ALWAYS, NEVER
            Whether to show seconds in clock 
        showHours : one of AUTO, ALWAYS, NEVER
            Whether to show hours in clock
        showDays : one of AUTO, ALWAYS, NEVER
            Whether to show days in clock
        allowReset : boolean
            If true, allow manual reset of all clocks
        '''
        super(ChessClockControl,self).__init__(# ChessclockControl
            doc,
            self.TAG,
            node              = node,
            name              = name,
            description       = description,
            buttonIcon        = buttonIcon,
            buttonText        = buttonText,
            buttonTooltip     = buttonTooltip,
            showHotkey        = showHotkey,
            pauseHotkey       = pauseHotkey,
            nextHotkey        = nextHotkey,
            startOpponentKey  = startOpponentKey,
            showTenths        = showTenths,
            showSeconds       = showSeconds,
            showHours         = showHours,
            showDays          = showDays,
            allowReset        = allowReset)
        print(node,addClocks)
        if node is not None or not addClocks:
            return
        
        print('--- Will add clocks')
        game   = self.getGame()
        roster = game.getPlayerRoster()[0]
        sides  = roster.getSides()
        for side in sides:
            name = side.getText()
            self.addClock(side        = name,
                          tooltip     = f'Clock for {name}',
                          buttonText  = name,
                          startHotkey = key('U'),
                          stopHotkey  = key('U'))
    
    def addClock(self,**kwargs):
        '''Add a clock element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : AboutScreen
            The added element
        '''
        return self.add(ChessClock,**kwargs)
    def getClocks(self,asdict=True):
        '''Return dictionary of clocs'''
        return self.getElementsByKey(ChessClock,'side',asdict)

registerElement(ChessClockControl)

#
# EOF
#



#!/usr/bin/env python
import sys
sys.path.append('..')

from common import Verbose, VerboseGuard
from common.collector import Collector

        
if __name__ == '__main__':
    from argparse import ArgumentParser, FileType  

    ap = ArgumentParser(description='Collect to single script')
    ap.add_argument('output',
                    type=FileType('w'),
                    nargs='?',
                    default='cyberboard.py',
                    help='Output script name')
    ap.add_argument('-v','--verbose',action='store_true',
                    help='Be verbose')

    args = ap.parse_args()

    Verbose().setVerbose(args.verbose)

    c = Collector(executable=False)
    c.run(args.output,
          '../common/singleton.py',
          '../common/verbose.py',
          '../common/verboseguard.py',
          'features.py',
          'archive.py',
          'base.py',
          'head.py',
          'image.py',
          'tile.py',
          'piece.py',
          'mark.py',
          'draw.py',
          'cell.py',
          'board.py',
          'gamebox.py',
          'scenario.py',
          'player.py',
          'windows.py',
          'palette.py',
          'tray.py',
          'extractor.py')

    



#!/usr/bin/env python
## BEGIN_IMPORTS
from pywargame.latex.latexexporter import LaTeXExporter
from pywargame.vassal.vmod import VMod
from pywargame.common import Verbose
## END_IMPORTS

from argparse import ArgumentParser

class DefaultSubcommandArgParse(ArgumentParser):
    _default_subparser = None

    def set_default_subparser(self, name):
        self._default_subparser = name

    def _parse_known_args(self, arg_strings, *args, **kwargs):
        from argparse import _SubParsersAction
        in_args = set(arg_strings)
        d_sp    = self._default_subparser
        if d_sp is not None and not {'-h', '--help'}.intersection(in_args):
            for x in self._subparsers._actions:
                subparser_found = (
                    isinstance(x, _SubParsersAction) and
                    in_args.intersection(x._name_parser_map.keys())
                )
                if subparser_found:
                    break
            else:
                # insert default in first position, this implies no
                # global options without a sub_parsers specified
                arg_strings = [d_sp] + arg_strings
        return super(DefaultSubcommandArgParse, self)._parse_known_args(
            arg_strings, *args, **kwargs
        )
# ====================================================================
def patchIt(args):
    vmodname  = args.output.name
    patchname = args.patch.name
    args.output.close()
    args.patch .close()
    
    VMod.patch(vmodname, patchname, args.verbose)

# ====================================================================
def exportIt(args):

    vmodname  = args.output.name
    patchname = args.patch.name if args.patch is not None else None

    args.output.close()
    if args.patch is not None:
        args.patch.close()

    Verbose().setVerbose(args.verbose)

    try:
        if args.version.lower() == 'draft':
            args.visible_grids = True
            
        rulesname = args.rules.name    if args.rules    is not None else None
        tutname   = args.tutorial.name if args.tutorial is not None else None
        
        exporter  = LaTeXExporter(vmodname      = vmodname,
                                  pdfname       = args.pdffile.name,
                                  infoname      = args.infofile.name,
                                  title         = args.title,
                                  version       = args.version,
                                  description   = args.description,
                                  rules         = rulesname,
                                  tutorial      = tutname,
                                  patch         = patchname,
                                  visible       = args.visible_grids,
                                  vassalVersion = args.vassal_version,
                                  nonato        = args.no_nato_prototypes,
                                  nochit        = args.no_chit_information,
                                  resolution    = args.resolution,
                                  counterScale  = args.counter_scale,
                                  imageFormat   = args.image_format)
        exporter.run()
    except Exception as e:
        from sys import stderr 
        print(f'Failed to build {vmodname}: {e}',file=stderr)
        from os import unlink
        try:
            unlink(vmodname)
        except:
            pass
        
        raise e
    

# ====================================================================
def exportMain():
    from argparse import ArgumentParser, FileType

    ap = DefaultSubcommandArgParse(description='Create draft VASSAL module')
    ap.set_default_subparser('export')
    sp = ap.add_subparsers(dest='mode')

    pp = sp.add_parser('patch',help='Patch VMod')
    pp.add_argument('output',
                    help='Module to patch',
                    type=FileType('r'),
                    default='Draft.vmod')
    pp.add_argument('patch',
                    help='A python script to patch generated module',
                    type=FileType('r'),
                    default='patch.py')
    pp.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')


    ep = sp.add_parser('export',help='Export from PDF and JSON to VMod')
    ep.add_argument('pdffile',
                    help='The PDF file to read images from',
                    type=FileType('r'),
                    default='export.pdf',
                    nargs='?')
    ep.add_argument('infofile',
                    help='The JSON file to read image information from',
                    type=FileType('r'),
                    default='export.json',
                    nargs='?')
    ep.add_argument('-o','--output',
                    help='Output file to write module to',
                    type=FileType('w'),
                    default='Draft.vmod')
    ep.add_argument('-p','--patch',
                    help='A python script to patch generated module',
                    type=FileType('r'))
    ep.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')
    ep.add_argument('-t','--title',
                    help='Module title', default='Draft',
                    type=str)
    ep.add_argument('-v','--version',
                    help='Module version',
                    type=str,
                    default='draft')
    ep.add_argument('-r','--rules',
                    help='Rules PDF file',
                    type=FileType('r'))
    ep.add_argument('-T','--tutorial',
                    help='Tutorial (v)log file',
                    type=FileType('r'))
    ep.add_argument('-d','--description',
                    help='Short description of module',
                    type=str,
                    default='draft of module')
    ep.add_argument('-W','--vassal-version',
                    help='Vassal version number',
                    type=str,
                    default='3.7.12')
    ep.add_argument('-G','--visible-grids',
                    action='store_true',
                    help='Make grids visible in the module')
    ep.add_argument('-N','--no-nato-prototypes',
                    action='store_true',
                    help='Do not make prototypes for types,echelons,commands')
    ep.add_argument('-C','--no-chit-information',
                    action='store_true',
                    help='Do not make properties from chit information')
    ep.add_argument('-S','--counter-scale',
                    type=float, default=1,
                    help='Scale counters by factor')
    ep.add_argument('-R','--resolution',
                    type=int, default=150,
                    help='Resolution of images')
    ep.add_argument('-I','--image-format',
                    choices = ['png','svg'], default='png',
                    help='Image format to use')
    
    args = ap.parse_args()
    
    if args.mode == 'patch':
        patchIt(args)
    else:
        exportIt(args)
    
# ====================================================================
if __name__ == '__main__':
    exportMain()
    
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class RestrictAccessTrait(Trait):
    ID = 'restrict'
    def __init__(self,
                 sides         = [],
                 byPlayer      = False,
                 noMovement    = True,
                 description   = '',
                 owner         = '',):
        '''Create a layer trait (VASSAL.counter.Restricted)'''
        super(RestrictAccessTrait,self).__init__()
        encSides = ','.join(sides)
        self.setType(sides         = encSides,
                     byPlayer      = byPlayer,
                     noMovement    = noMovement,
                     description   = description)
        self.setState(owner=owner)

Trait.known_traits.append(RestrictAccessTrait)

#
# EOF
#



## BEGIN_IMPORTS
from . base import DomInspect, ZTImage
from . piece import ZTPiece
from pywargame.common import VerboseGuard
## END_IMPORTS

# --------------------------------------------------------------------
class ZTCounterImage(ZTImage):
    def __init__(self,elem,prefix):
        super().__init__(elem,prefix)
        self._mask_file = self._get_attr(elem,prefix+'mask-file')

    def read_image(self,zf):
        from wand.image import Image
        
        super().read_image(zf)
        
        self._mask = self._read_image(zf,self._mask_file,self._reso)
        if not self._mask:
            return

        
        # Get blue channel and make alpha mash from that 
        blue               = self._mask.channel_images['blue']
        blue.alpha_channel = 'copy'

        # Make copy of input image, and turn on alpha channel for that
        cp                 = Image(image=self._image)
        cp.alpha_channel   = 'set'

        # Compose with alpha of blue mask
        cp.composite(blue,operator='dst_in')

        # Set new image (now with alpha channel)
        self._orig  = self._image
        self._image = cp
        self._blue  = blue
 
# --------------------------------------------------------------------
class Face(DomInspect):
    def __init__(self,elem,prefix=''):
        with VerboseGuard(f'Got a face'):
            self._left    = int(self._get_attr(elem,prefix+'left',  -1))
            self._top     = int(self._get_attr(elem,prefix+'top',   -1))
            self._right   = int(self._get_attr(elem,prefix+'right', -1))
            self._bottom  = int(self._get_attr(elem,prefix+'bottom',-1))

    @property
    def width(self):
        return self._right - self._left

    @property
    def height(self):
        return self._bottom - self._top

    def get_tiles(self,image,dx,dy,reverse):
        with VerboseGuard(f'Get tiles from image {dx},{dy},{reverse}') as v:
            # Check ordering - ZunTzu goes row-wise first
            xl = self._left  if not reverse else self._right-dx
            xh = self._right if not reverse else self._left -dx
            x1 = self._left
            x2 = self._right
            yl = self._top
            yh = self._bottom
            sx = dx          if not reverse else -dx
            
            def crop(image,x,y):
                xx  = max(x,   x1)
                xxx = min(x+dx,x2)
                yy  = max(y,   yl)
                yyy = min(y+dy,yh)
                if xxx-xx < 0 or yyy-yy < 0:
                    return
                return image[xx:xxx,yy:yyy],x+dx//2,y+dy//2
            
            v(f'{x1} <= x+dx <= {x2} {yl} <= y+dy <= {yh}')
            return [crop(image,x,y)
                    for y in range(yl, yh-dy//2, dy)
                    for x in range(xl, xh-sx//2, sx)
                    ]

    def __str__(self):
        return f'Face: {self._left},{self._top},{self._right},{self._bottom}'
        
        
# --------------------------------------------------------------------
class Section(DomInspect):
    def __init__(self,elem,terrain):
        with VerboseGuard(f'Got an image section') as v:
            self._terrain = terrain
            self._card    = elem.tagName == 'card-section'
            self._type    = self._get_attr(elem,'type',0)
            self._rows    = int(self._get_attr(elem,'rows',1))
            self._cols    = int(self._get_attr(elem,'columns',1))
            self._shadow  = float(self._get_attr(elem,'shadow',
                                                 0 if terrain else 20))
            self._supply  = int(self._get_attr(elem,'supply',1))
            self._front   = Face(elem,'face-' if self._card else 'front-')
            self._back    = Face(elem,'back-')
            if self._terrain:
                self._shadow = 0;
                
            v(f'{"Terrain" if self._terrain else "Piece"} '
              f'{"Card" if self._card else "Piece"} '
              f'{self._type} {self._rows}x{self._cols} {self._supply}')

    @property
    def expected(self):
        return self._rows * self._cols * self._supply
    
    def get_tiles(self,image,face,reverse):
        with VerboseGuard(f'Get tiles from face {face}') as v:
            dx = int(face.width / self._cols+.5)
            dy = int(face.height / self._rows+.5)
            #v(f'Offsets are {dx},{dy}')
            if dx <= 0 or dy <= 0:
                return [(None,None,None)]*(self._cols*self._rows)

        return face.get_tiles(image,dx,dy,reverse)

    def make_pieces(self,
                    front_image,
                    back_image,
                    front_reso,
                    back_reso):
        with VerboseGuard(f'Make pieces from images {self._supply}. '
                          f'Expecting {self.expected}') as v:
            fronts = self.get_tiles(front_image, self._front, False)
            backs  = self.get_tiles(back_image,  self._back,  True)

            fs = ZTImage.target_dpi / front_reso
            bs = ZTImage.target_dpi / back_reso
            #v(f'Scales are {fs} and {bs} '
            #  f'({ZTImage.target_dpi} {front_reso},{back_reso})')
            
            def fixup(i,f,s):
                if not i:
                    return

                i.resize(int(f * i.width),int(f * i.height))
                c = i.clone()
                o = min(min(i.width,i.height) * s / 100 / 5,8)
                c.shadow(50,o,0,0)
                c.negate(channel='rgb')
                c.composite(i,int(o/4),int(o/4))
                return c
            
            def make(f,b):
                fi, fx, fy = f
                bi, bx, by = b
                #v(f' fx={fx} fy={fy} bx={bx} by={by}')
                x          = fs * fx if f else bs * bx
                y          = fs * fy if f else bs * by
                fi         = fixup(fi,fs,self._shadow)
                bi         = fixup(bi,bs,self._shadow)
                #v(f' x={x} y={y} w={fi.width} h={fi.height}')
                return [ZTPiece(fi,bi,int(x),int(y),self._terrain,self._card)
                        for _ in range(self._supply)]
        
            return sum([make(f,b) for f,b in zip(fronts,backs)],[])
            
    def __str__(self):
        return (f'  Section: {"terrain" if self._terrain else "piece"} '+
                f'{"card" if self._card else "piece"} '+
                f'{self._rows}x{self._cols} (x{self._supply}) '+
                f'type={self._type} shadow={self._shadow}%'+'\n'
                f'   Front: '+str(self._front)+'\n'
                f'   Back:  '+str(self._back))
        
# --------------------------------------------------------------------
class ZTCounterSheet(DomInspect):
    def __init__(self,elem):
        self._name    = self._get_attr(elem,'name')
        with VerboseGuard(f'Got a counter sheet "{self._name}"'):
            self._terrain = elem.tagName == 'terrain-sheet'
            self._name    = self._get_attr(elem,'name')
            self._front   = ZTCounterImage(elem,'front-')
            self._back    = ZTCounterImage(elem,'back-')

            self._sections = [self.parse_section(sc)
                              for sc in self._find_children(elem,
                                                            'counter-section',
                                                            'terrain-section')]
            self._cards    = [self.parse_card(cd)
                              for cd in self._find_children(elem,
                                                            'card-section')]
        
    def parse_section(self,section):
        return Section(section,self._terrain)


    def parse_card(self,card):
        return Section(card,self._terrain)
    
    def read_image(self,zf):
        self._front.read_image(zf)
        self._back .read_image(zf)

    def make_pieces(self):
        with VerboseGuard(f'Making pieces from counter sheet {self._name}') \
             as v:
            self._piece = sum([s.make_pieces(self._front._image,
                                             self._back._image,
                                             self._front._reso,
                                             self._back._reso)
                               for s in self._sections],[])
            self._card  = sum([c.make_pieces(self._front._image,
                                             self._back._image,
                                             self._front._reso,
                                             self._back._reso)
                               for c in self._cards],[])
            exp_piece = sum([s.expected for s in self._sections])
            exp_card  = sum([s.expected for s in self._cards])
            v(f'{len(self._piece)} pieces and {len(self._card)} cards, '
              f'expected {exp_piece} pieces and {exp_card} cards')

            for img, res in zip([self._front._image,self._back._image],
                                [self._front._reso, self._back._reso]):
                if not img:
                    continue

                s = ZTImage.target_dpi / res
                #v(f'scale image from {img.width}x{img.height} by {s}')
                img.resize(int(s*img.width+.5),
                           int(s*img.height+.5))

    @property
    def filename(self):
        return f'{self._name.replace(" ","_")}.{self._front._image.format.lower()}'

    @property
    def size(self):
        return self._front._image.width,self._front._image.height
        
    def __str__(self):
        return (f' Sheet: {self._name} '+
                f'{"terrain" if self._terrain else "pieces"} '+'\n'+
                f'  Front: '+str(self._front)+'\n'+
                f'  Back:  '+str(self._back)+
                ('\n' if len(self._sections)>0 else '')+
                '\n'.join([str(s) for s in self._sections])+
                ('\n' if len(self._cards)>0 else '')+
                '\n'.join([str(s) for s in self._cards])+
                ('\n' if len(self._piece)>0 else '')+
                '\n'.join([str(p) for p in self._piece])+
                ('\n' if len(self._card)>0 else '')+
                '\n'.join([str(c) for c in self._card]))
#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . head    import *
from . player  import GSNPlayerManager
from . windows import GSNWindows
from . palette import GSNTrayPalette
from . palette import GSNMarkPalette
from . board   import GSNBoardManager
from . tray    import GSNTrayManager
from . piece   import GSNPieceTable
from . gamebox import GameBox
## END_IMPORT

class GSNInfo:
    def __init__(self,ar):
        '''Scenario information'''
        self._disableOwnerTips = ar.word()
        self._reserved         = [ar.word() for _ in range(3)]
        self._gameID           = ar.dword()  # 4 -> 14
        self._majorRevs        = ar.dword()  # 4 -> 6
        self._minorRevs        = ar.dword()  # 4 -> 10
        self._gbxFilename      = ar.str()
        self._scenarioID       = ar.dword()
        self._title            = ar.str()    # Y  -> 30+X+Y
        self._author           = ar.str()    # X  -> 30+X
        self._description      = ar.str()    # Z  -> 30+X+Y+Z
        self._keepBackup       = ar.word()
        self._keepHistory      = ar.word()
        self._verifyState      = ar.word()
        self._verifySave       = ar.word()
        self._showObjectTip    = ar.word()

    def __str__(self):
        return ('Information:\n'
                f'  Disable owner tips: {self._disableOwnerTips}\n'
                f'  Reserved:           {self._reserved}\n'
                f'  Game ID:            {self._gameID}\n'
                f'  Major revision:     {self._majorRevs}\n'
                f'  Minor revision:     {self._minorRevs}\n'
                f'  Game box filename:  {self._gbxFilename}\n'
                f'  Scenario ID:        {self._scenarioID}\n'
                f'  Title:              {self._title}\n'
                f'  Author:             {self._author}\n'
                f'  Description:        {self._description}\n'
                f'  Keep backup:        {self._keepBackup}\n'
                f'  Keep history:       {self._keepHistory}\n'
                f'  Verify state:       {self._verifyState}\n'
                f'  Verify save:        {self._verifySave}\n'
                f'  Show object tips:   {self._showObjectTip}\n')
                
# ====================================================================
class Scenario:
    def __init__(self,ar,gbxfilename=None):
        '''Container of game'''
        with VerboseGuard(f'Reading scenario'):
            self._header        = GBXHeader       (ar,GBXHeader.SCENARIO)
            self._info          = GSNInfo         (ar)
            self._strings       = GSNStrings      (ar)
            self._playerManager = GSNPlayerManager(ar)
            self._windows       = GSNWindows      (ar)
            self._trayA         = GSNTrayPalette  (ar,self._header._vers,'A')
            self._trayB         = GSNTrayPalette  (ar,self._header._vers,'B')
            self._mark          = GSNMarkPalette  (ar,self._header._vers)
            self._boards        = GSNBoardManager (ar,self._header._vers)
            self._trayManager   = GSNTrayManager  (ar,self._header._vers)
            self._pieceTable    = GSNPieceTable   (ar,self._header._vers)
            # Possibly override GBX file name 
            if gbxfilename is not None and gbxfilename != '':
                self._info._gbxFilename = gbxfilename

            self.readGameBox(ar)

    def readGameBox(self,ar):
        from pathlib import Path
        with VerboseGuard(f'Read game box file {self._info._gbxFilename}') as v:
            gbxname = self._info._gbxFilename
            gbxpath = Path(gbxname)
            
            if not gbxpath.exists():
                v(f'GameBox file {gbxpath} does not exist')
                if '\\' in gbxname:
                    gbxname = gbxname.replace('\\','/')
                    gbxpath = Path(gbxname)
                gbxpath = ar.path.parent / Path(gbxpath.name)
                if not gbxpath.exists():
                    raise RuntimeError(f'GameBox file {gbxpath} cannot be found')
                
            self._gbx = GameBox.fromFile(str(gbxpath))

            if self._gbx._info._gameID != self._info._gameID:
                raise RuntimeError(f'Game IDs from GBX and GSN does not match: '
                                   f'{self._gbx._info._gameID} versus '
                                   f'{self._header._gameID}')
        
    def __str__(self):
        return (str(self._header)+      
                str(self._info)+        
                str(self._strings)+
                str(self._playerManager)+
                str(self._windows)+
                str(self._trayA)+
                str(self._trayB)+
                str(self._mark)+
                str(self._boards)+
                str(self._trayManager)+
                str(self._pieceTable)+
                str(self._gbx)+
                ''
                )

    
    @classmethod
    def fromFile(cls,filename,gbxfile=None):
        with Archive(filename,'rb') as ar:
            return Scenario(ar,gbxfile)



# --------------------------------------------------------------------
## BEGIN_IMPORT
from . verbose import Verbose
## END_IMPORT

class VerboseGuard:
    def __init__(self,*args,**kwargs):
        '''A guard pattern that increases verbose indention

        This is a context manager.  The arguments passed are used for
        an initial message, before increasinig indention.

        Parameters
        ----------
        args : tuple
            Arguments
        kwargs : dict
            Keyword arguments
        '''                
        Verbose().message(*args,**kwargs)

    def __bool_(self):
        '''Test if verbose'''
        return Verbose().verbose
    
    def __enter__(self):
        '''Enter context'''
        Verbose().incr()
        return self

    def __exit__(self,*args):
        '''Exit context'''        
        Verbose().decr()

    @property
    def i(self):
        return Verbose()._indent
    
    def __call__(self,*args,**kwargs):
        '''Write a message at current indention level
        
        Parameters
        ----------
        args : tuple
            Arguments
        kwargs : dict
            Keyword arguments
        '''                
        Verbose().message(*args,**kwargs)

#
# EOF
#



# --------------------------------------------------------------------
## BEGIN_IMPORT
from . singleton import Singleton
## END_IMPORT

class Verbose(metaclass=Singleton):
    def __init__(self,verbose=False):
        '''Singleton for writing message to screen, contigent on setting

        Parameters
        ----------
        verbose : bool
             Whether to show messages or not
        '''
        self._indent  = ''
        self._verbose = verbose

    def setVerbose(self,verbose):
        '''Set whether to print or not

        Parameters
        ----------
        verbose : bool
             Whether to show messages or not
        '''
        self._verbose = verbose

    @property
    def verbose(self):
        '''Test if this is verbose'''
        return self._verbose

    def message(self,*args,**kwargs):
        '''Write messsage if verbose

        Parameters
        ----------
        args : tuple
            Arguments
        kwargs : dict
            Keyword arguments
        '''        
        if not self._verbose: return
        if not kwargs.pop('noindent', False):
            print(self._indent,end='')
        print(*args,**kwargs)

    def incr(self):
        '''Increment indention'''
        self._indent += ' '

    def decr(self):
        '''Decrement indention'''
        if len(self._indent) > 0:
            self._indent = self._indent[:-1]

#
# EOF
#



#
#
# 
from setuptools import setup

def get_version():
    from os import getenv

    for e in ['CI_COMMIT_REF_NAME','CI_JOB_ID']:
        v = getenv(e,False)
        if v:
            return {'version':str(v)}

    return {}

setup(**get_version())

#
# EOF
#
